from pathlib import Path as pathlib_Path

def dummy_function(*args, **kwargs): ...

class DummyClass:
    def __init__(self, *args, **kwargs):
        print('warning: this is a dummy class! It returns '
              'None to whatever you throw at it.')

    def __getattr__(self, name):
        return self.dummy_function

    def __call__(self, *args, **kwargs):
        return None

    def dummy_function(self, *args, **kwargs):
        return None

def is_builtin_collection(obj):
    """
    Determine if an object is a strictly built-in Python collection.
    
    This function uses a heuristic based on the object 
    type's module being either 'builtins',
    'collections', or 'collections.abc', excluding 
    strings and bytes explicitly, to identify
    if the given object is a built-in collection 
    type (list, tuple, dict, set). It checks if the
    object belongs to one of Python's built-in 
    collection modules and possesses both __len__ and
    __iter__ methods, which are typical characteristics of collections.
    
    Args:
        obj: The object to be checked.
    
    Returns:
        bool: True if the object is a built-in Python 
        collection (excluding strings and bytes),
              False otherwise.
    
    Note:
        This function aims to exclude objects from external 
        libraries (e.g., NumPy arrays) that,
        while iterable and having a __len__ method, 
        are not considered built-in Python collections.
    """
    obj_type = type(obj)
    module = obj_type.__module__
    if ( (module not in ('builtins', 'collections', 'collections.abc'))
         | isinstance(obj, (str, bytes)) 
        ):
        return False
    return hasattr(obj, '__len__') and hasattr(obj, '__iter__')

def assure_is_collection(returned_obj):
    if not is_builtin_collection(returned_obj):
        return [returned_obj]
    return returned_obj

def name_from_file(log_dir, fpath):
    """ 
        Given an fpath inside the logger log_dir, 
        what would be its equivalent parameter_name?
    """
    fpath_str = str(fpath.absolute())
    try:
        log_dir = str(log_dir.absolute())
    except:
        log_dir = str(log_dir)
    log_dir_str = None
    if log_dir in fpath_str:
        log_dir_str = log_dir
    if (log_dir + '/') in fpath_str:
        log_dir_str = log_dir + '/'
    if log_dir_str:
        fpath_name = fpath_str.split(log_dir_str)[-1]
        fpath_split = fpath_name.split('.')
        return '.'.join(fpath_split[:-1])
    
def repr_raw(text):
    """ Raw text representation
        Returns a raw string representation of a text that has escape 
        charachters
        
        Parameters:
        ^^^^^^^^^
        :param text:
        the input text, returns the fixed string
        
    """
    escape_dict={r'\a':r'\a',
                 r'\b':r'\b',
                 r'\c':r'\c',
                 r'\f':r'\f',
                 r'\n':r'\n',
                 r'\r':r'\r',
                 r'\t':r'\t',
                 r'\v':r'\v',
                 r'\'':r'\'',
                 r'\"':r'\"'}
    new_string=''
    for char in text:
        try: 
            new_string += escape_dict[char]
        except KeyError: 
            new_string += char
    return new_string

def replace_all(text, pattern, fill_value):
    """replace all instances of a pattern in a string with a new one
    """
    while (len(text.split(pattern)) > 1):
        text = text.replace(pattern, fill_value)
    return text

def select_directory(default_directory = './'):
    """ Open dialog to select a directory
        It works for windows and Linux using PyQt5.
    
       :param default_directory: pathlib_Path
                When dialog opens, it starts from this default directory.
    """
    from PyQt5.QtWidgets import QFileDialog, QApplication
    _ = QApplication([])
    log_dir = QFileDialog.getExistingDirectory(
        None, "Select a directory", default_directory, QFileDialog.ShowDirsOnly)
    return(log_dir)

def select_file():
    """ Open dialog to select a file
        It works for windows and Linux using PyQt5.
    """
    from PyQt5.QtWidgets import QFileDialog, QApplication
    _ = QApplication([])
    fpath = QFileDialog.getOpenFileName()
    fpath = pathlib_Path(fpath[0])
    return(fpath)

def text_to_collection(text):
    """ Read a list or dict that was sent to write to text e.g. via log_single:
    As you may have tried, it is possible to send a Pythonic list to a text file
    the list will be typed there with [ and ] and ' and ' for strings with ', '
    in between. In this function we will merely return the actual content
    of the original list.
    Now if the type the element of the list was string, it would put ' and ' in
    the text file. But if it is a number, no kind of punctuation or sign is 
    used. by write(). We support int or float. Otherwise the written text
    will be returned as string with any other wierd things attached to it.
    
    """
    import ast
    def parse_node(node):
        if isinstance(node, ast.List):
            return [parse_node(elem) for elem in node.elts]
        elif isinstance(node, ast.Dict):
            return {parse_node(key): parse_node(value) 
                    for key, value in zip(node.keys, node.values)}
        elif isinstance(node, ast.Constant):
            return node.value
        elif isinstance(node, ast.Num):  # For Python < 3.8
            return node.n
        elif isinstance(node, ast.Str):  # For Python < 3.8
            return node.s
        elif isinstance(node, ast.Name):
            if node.id == 'array':
                import numpy
                return numpy
            elif node.id == 'tensor':
                import torch
                return torch
        elif isinstance(node, ast.Call):
            func_name = node.func.id
            if func_name == 'array':
                import numpy as np
                return np.array([parse_node(arg) for arg in node.args])
            elif func_name == 'tensor':
                import torch
                return torch.tensor([parse_node(arg) for arg in node.args])
        return None

    tree = ast.parse(text, mode='eval')
    return parse_node(tree.body)

class SSHSystem:
    """
    A class to handle basic SSH and SFTP operations on a remote system.

    Attributes:
        ssh_client (paramiko.SSHClient): The SSH client for executing 
        commands on the remote system.
        sftp_client (paramiko.SFTPClient): The SFTP client for file 
        transfer operations.
    """

    def __init__(self, hostname: str, username: str, password: str):
        """
        Initialize the SSHSystem by setting up the SSH and SFTP clients.

        Args:
            hostname (str): The hostname or IP address of the remote system.
            username (str): The username for SSH authentication.
            password (str): The password for SSH authentication.
        """
        import paramiko
        self.ssh_client = paramiko.SSHClient()
        self.ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            self.ssh_client.connect(
                hostname=hostname, username=username, password=password)
            self.sftp_client = self.ssh_client.open_sftp()
        except Exception as e:
            print(f"Failed to connect to {hostname}: {e}")
            self.ssh_client = None
            self.sftp_client = None

    def ssh_ls(self, path: pathlib_Path):
        """
        List the contents of a directory on the remote system.

        Args:
            path (pathlib_Path): The path to the directory on the remote system.

        Returns:
            list: A list of pathlib_Path objects representing the files in the directory.
        """
        try:
            stdin, stdout, stderr = self.ssh_client.exec_command(f'ls {path}')
            ls_result = stdout.readlines()
            return [path / file.strip() for file in ls_result]
        except Exception as e:
            print(f"Error listing directory {path}: {e}")
            return []

    def ssh_scp(self, source: pathlib_Path, destination: pathlib_Path):
        """
        Copy a file from the remote system to the local system using SFTP.

        Args:
            source (pathlib_Path): The path of the file on the remote system.
            destination (pathlib_Path): The path where the file will be saved locally.
        """
        try:
            self.sftp_client.get(str(source), str(destination))
        except Exception as e:
            print(f"Error copying {source} to {destination}: {e}")

    def ssh_rm(self, path: pathlib_Path):
        """
        Remove a file from the remote system.

        Args:
            path (pathlib_Path): The path to the file to be removed.

        Returns:
            tuple: A tuple containing the stdout and stderr outputs from the command.
        """
        try:
            stdin, stdout, stderr = self.ssh_client.exec_command(f'rm {path}')
            return stdout.read().decode(), stderr.read().decode()
        except Exception as e:
            print(f"Error removing file {path}: {e}")
            return "", str(e)

    def monitor_and_move(
        self, remote_folder: pathlib_Path, target_fname: str,
        local_folder: pathlib_Path, interval=30
    ):
        """
        Monitor a remote folder for a specific file. Once the file appears,
        transfer and delete other files from the folder.

        Args:
            remote_folder (pathlib_Path): The folder on the remote system to monitor.
            local_folder (pathlib_Path): The local folder where files will be copied.
            target_fname (str): The name of the file to wait for.
            interval (int, optional): The time interval (in seconds) 
            between each check. Default is 30 seconds.
        """
        interesting_file_path = remote_folder / target_fname
        cnt = 0
        import time
        while not self.is_file(interesting_file_path):
            if (cnt % 100) == 0:
                print(f'Waiting for {interesting_file_path}', end='')
            else:
                print('.', end='', flush=True)
            time.sleep(interval)
            cnt += 1
        print('')

        print(f"{target_fname} found! Starting file transfer and deletion.")
        files = self.ssh_ls(remote_folder)
        for file in files:
            local_file = local_folder / file.name
            
            # Copy file to local folder
            print(f"Copying {file} to {local_file}")
            self.ssh_scp(file, local_file)
            
            # Delete file from remote server
            print(f"Deleting {file} from remote server")
            self.ssh_rm(file)

    def is_file(self, path: pathlib_Path) -> bool:
        """
        Check if a file exists on the remote system.

        Args:
            path (pathlib_Path): The path to the file on the remote system.

        Returns:
            bool: True if the file exists, False otherwise.
        """
        try:
            stdin, stdout, stderr = self.ssh_client.exec_command(
                f'test -f {path} && echo "exists"')
            return "exists" in stdout.read().decode()
        except Exception as e:
            print(f"Error checking file {path}: {e}")
            return False

    def close_connection(self):
        """
        Close the SSH and SFTP connections to the remote system.
        """
        if self.sftp_client:
            self.sftp_client.close()
        if self.ssh_client:
            self.ssh_client.close()

def printv(var, logger = print, tab = 0,
           arr_size_max = 1e6, arr_size_min = 50,
           str_len_max = 1e4, str_len_min = 50):
    """printv
    Provides a detailed description of a variable, including its type, size, and basic statistics,
    and logs the output using a specified logger. This function is particularly useful for inspecting
    variables in machine learning workflows and large data structures.

    Parameters:
    ----------
    var : any
        The variable to be logged and described.
    logger : callable, optional
        A function to output the description, such as `print` or a logging function. Default is `print`.
    tab : int, optional
        The indentation level for nested structures. Default is 0.
    arr_size_max : int, optional
        The maximum number of elements in an array-like structure to print its statistics
        (min, max, mean, std). Default is 1e6.
    arr_size_min : int, optional
        if the size of a numpu or torch array is less than this we just print it.
    str_len_max : float, optional
        The maximum character length for displaying string representations of the variable.
        Beyond this length, only a truncated representation is shown. Default is 1e4.
    str_len_min : float, optional
        The minimum character length for displaying string representations of the variable.
        Below this length, the str of the parameter is shown. Default is 50.
    Returns:
    -------
    int
        The length of the logged description string.

    Notes:
    ------
    - For `int`, `float`, or `bool` types, the function logs the value directly.
    - For arrays (NumPy or PyTorch), it logs the shape, data type, device, and optionally
      basic statistics (if array size is below `arr_size_max`).
    - For lists, tuples, and dictionaries, the function logs the length and recursively
      inspects each element up to `str_len_max`.
    - For strings and other data types, it truncates and displays the string if it's too long.
    """
    
    import inspect
    frame = inspect.currentframe().f_back
    var_name = [name for name, value in frame.f_locals.items() if value is var]
    
    if var_name:
        var_name = var_name[0]
    else:
        var_name = repr(var)

    toprint = f'{var_name}: {type(var).__name__}'
    
    if (isinstance(var, int)) | (isinstance(var, float)) | (isinstance(var, bool)):
        toprint += ', ' + str(var)
        logger(toprint)
        return len(toprint)
    
    has_element_0 = False
    try:
        var_ = var.item()
        has_element_0 = True
    except: pass
    if has_element_0:
        try:
            var[1]
        except:
            toprint += ', ' + str(var_)
            logger(toprint)
            return len(toprint)
    try:
        array_shape = var.shape
        is_np_or_torch = True
    except: 
        is_np_or_torch = False
    if is_np_or_torch:
        import numpy as np
        toprint += f', shape={array_shape}'
        try:
            array_dtype = var.dtype
            toprint += f', dtype={array_dtype}'
        except: pass
        try:
            toprint += f', device={var.device}'
        except: pass
        
        arr_size = np.prod(array_shape)
        if arr_size < arr_size_max:
            try:
                toprint += f', min={var.min():.6f}'
            except: pass
            try:
                toprint += f', max={var.max():.6f}'
            except: pass
            try:
                toprint += f', mean={var.mean():.6f}'
            except: pass
            try:
                toprint += f', std={var.std():.6f}'
            except: pass
        if arr_size < arr_size_min:
            var_str = str(var)
            if len(var_str) < str_len_min:
                toprint += '\n' + var_str + ' <--' + var_name
            
        logger(toprint)
        return len(toprint)

    if is_builtin_collection(var):
        toprint += f', len={len(var)}'
        var_str = str(var)
        if len(var_str) < str_len_min:
            toprint += ', ' + var_str
            logger(toprint)
            return len(toprint)
        if isinstance(var, (list, tuple)):
            logger(toprint)
            len_toprint = 0
            tab += 1
            for idx, item in enumerate(var):
                logger('    ' * tab + f"{var_name}[{idx}]: ", end = '')
                len_toprint += printv(item, logger, tab = tab,
                                      arr_size_max = arr_size_max,
                                      str_len_max = str_len_max)
                if len_toprint > str_len_max:
                    logger(f"... too long")
                    break
            return len(toprint) + len_toprint
        elif isinstance(var, dict):
            logger(toprint)
            len_toprint = 0
            tab += 1
            for key, item in var.items():
                logger('    '*tab + f"{var_name}[{repr(key)}]: ", end = '')
                len_toprint += printv(item, logger, tab = tab,
                                      arr_size_max = arr_size_max,
                                      str_len_max = str_len_max)
                if len_toprint > str_len_max:
                    logger(f"... too long")
                    break
            return len(toprint) + len_toprint
            
    var_str = str(var)
    if len(var_str) > str_len_max:
        toprint += ', too long'
        var_str = var_str[:int(str_len_max // 10)] + ' ...' + \
            var_str[-int(str_len_max // 10):]
    toprint += ', ' + var_str
    logger(toprint)
    return len(toprint)

def find_duplicates(a_list):
    from collections import Counter
    element_count = Counter(a_list)
    duplicates = [item for item, count in element_count.items() if count > 1]
    return duplicates

class block_runner:
    """
    A Jupyter-like Python code runner that executes code in blocks based 
    on cell numbers, supports saving and loading kernel states, and allows 
    interactive execution.

    Attributes:
        fpath (pathlib_Path): The path to the Python file to execute.
        logger_ (callable): An optional logger function to log messages.
        log (str): A string containing the accumulated log messages.
        saved_state (dict): A dictionary to hold saved kernel states.
        exit (bool): A flag to indicate when to stop execution.
    """

    def __init__(self, fpath: str, logger=None, figsize = None,
                 block_identifier = 'code_block_id'):
        """
        Initializes the block_runner class, runs the Python file in an interactive loop,
        and allows execution of specific code blocks identified by cell numbers.

        Args:
            fpath (str): The file path to the Python script to be executed.
            logger (callable, optional): A logger function to log output 
            (default is None).
            block_identifier: string that block_runner will be looking for in your
            code to find blocks of code to run. So you must struction you code
            to have blocks of code separated using if block_identifier == a_number:
                e.g.:
                if block_identifier == 0:
                    do_this()
                if block_identifier == 1:
                    do_that()
            
        """
        import runpy
        self.figsize = figsize
        self.block_identifier = block_identifier
        self.logger_ = logger
        self.fpath = pathlib_Path(fpath)
        assert self.fpath.is_file(), f"File {fpath} does not exist."
        self.log = ''
        self.saved_state = {}
        self.exit = False

        self.logger(f'file: {fpath}')
        while not self.exit:        
            show_and_ask_result = self.show(globals())
            if show_and_ask_result is None:
                continue
            globals().update(show_and_ask_result)
            globals().update({"__name__": "__main__"})
            try:
                # exec(globals().get('block_runner_code', ''), globals())
                result = runpy.run_path(fpath, init_globals=globals())
                globals().update(result)
                
            except Exception as e:
                print('-----------Block runner error-------------')
                print(e)
                print('-'*40)

    def logger(self, toprint: str, end: str = '\n'):
        """
        Logs the provided message. If a logger is provided, it logs the message 
        using that function. Otherwise, it appends the message to the internal log.

        Args:
            toprint (str): The message to log.
            end (str, optional): The string appended after each message 
            (default is '\n').
        """
        toprint = str(toprint) + end
        self.log += toprint
        if self.logger_ is not None:
            self.logger_(toprint)

    def save_or_load_kernel_state(self, globals_: dict, saved_state=None):
        """
        Saves or loads the kernel state using the `dill` library. 
        If `saved_state` is provided, it loads
        the state into `globals_`. If `saved_state` is None, it returns 
        a serialized form of the current global variables.

        Args:
            globals_ (dict): The global variables to save or update.
            saved_state (bytes, optional): The serialized kernel state to load 
            (default is None).

        Returns:
            bytes: A serialized version of the global variables if saving the state.
        """
        import dill as pickle
        if saved_state is None:
            return pickle.dumps(
                {k: v for k, v in globals_.items() 
                 if not k.startswith('__') and not callable(v)}
            )
        else:
            globals_.update(pickle.loads(saved_state))

    @property
    def n_saves(self) -> int:
        """
        Returns the number of saved states.

        Returns:
            int: The number of saved states.
        """
        return len(self.saved_state.keys())

    def show(self, globals_: dict) -> dict:
        """
        Displays available cell blocks for execution and handles user 
        interaction to run specific blocks or manage kernel states 
        (save/load/delete).

        Args:
            globals_ (dict): The global variables of the current session.
            figsize (tuple, optional): The size of the dialog box 
            (default is (3, 2)).

        Returns:
            dict: A dictionary containing the updated global variables 
            if a cell block is selected.
        """
        block_runner_code = open(self.fpath).read()
        
        # Updated regex to match both numbers (integers, floats) and text identifiers
        pattern = rf"if\s+{self.block_identifier}\s*==\s*(.+?):"
        import re
        matches = re.findall(pattern, block_runner_code)
        
        if len(matches) == 0:
            self.logger(f'Running the block_runner_code in {self.fpath}')
            self.logger(f'No code blocks found that checks {self.block_identifier}')
            return
        
        block_identifiers = []
        for item in matches:
            is_number = False
            try:
                _item = float(item)
                is_number = True
            except: pass
                
            if is_number:
                if _item == int(_item):
                    _item = int(_item)
            else:
                _item = str(item.strip("'").strip('"'))
            block_identifiers.append(_item)
        for blk in block_identifiers:
            if isinstance(blk, str):
                if ('load' in blk) | ('save' in blk) | ('exit' in blk):
                    self.logger(
                        f'block identifier {blk} is using a preserved word.'
                        ' Please do not use load, save and exit to name your'
                        'code blocks.')
        for blk_id_rep in find_duplicates(block_identifiers):
            self.logger(f'block identifier {blk_id_rep} is used more than once.')
        buttons = {}
        for block_identifier in block_identifiers:
            buttons[f'{block_identifier}'] = block_identifier
            
        # Add options for saved states
        for key in self.saved_state:
            buttons[f'load_{key}'] = f'load_{key}'
            buttons[f'del_{key}'] = f'del_{key}'

        buttons[f'save_{self.n_saves + 1}'] = f'save_{self.n_saves + 1}'
        buttons['exit'] = 'exit'

        # Display dialog for user interaction
        from lognflow.plt_utils import question_dialog
        show_and_ask_result = question_dialog(
            question='Choose a cell number', figsize=self.figsize, buttons=buttons
        )
        if show_and_ask_result is None:
            self.logger(f'block_runner: closing reloads, press Exit to close.')
            return

        # Handle user selection
        if isinstance(show_and_ask_result, str):
            if show_and_ask_result == 'exit':
                self.exit = True
                return

            elif 'save' in show_and_ask_result:
                key = show_and_ask_result.split('save_')[1]
                self.saved_state[key] = self.save_or_load_kernel_state(globals_)
                self.logger(f'Saved state: {key}')
                return

            elif 'load' in show_and_ask_result:
                key = show_and_ask_result.split('load_')[1]
                self.save_or_load_kernel_state(globals_, self.saved_state[key])
                self.logger(f'Loaded state: {key}')
                return

            elif 'del' in show_and_ask_result:
                key = show_and_ask_result.split('del_')[1]
                self.saved_state.pop(key)
                self.logger(f'Deleted state: {key}')
                return

        globals_['block_runner_code'] = block_runner_code
        globals_[self.block_identifier] = show_and_ask_result
        return globals_




#!/usr/bin/env python

"""Tests for `lognflow` package."""
import pytest
import time
import inspect
import matplotlib.pyplot as plt
import lognflow
from lognflow.plt_utils import (
    plt_imshow, complex2hsv_colorbar, plt_imhist,complex2hsv,
    transform3D_viewer, plt_mark, plt_contours, question_dialog,
    plt_plot, plt_hist2, plt_confusion_matrix, pv_volume, pv_surface,
    plt_imshow_subplots)
import numpy as np

def test_plt_imshow_subplots_comples():
    data = [np.random.rand(100, 100) + 1j * np.random.rand(100, 100),
            np.random.rand(100, 100), np.random.rand(100, 100),
            np.random.rand(100, 100)]
    
    plt_imshow_subplots(data, colorbar = True, vmin = 0.3, vmax = 0.4)
    plt.show()

def test_transform3D_viewer():
    print('Testing function', inspect.currentframe().f_code.co_name)
    in_pointcloud = np.random.randn(100, 3)
    moving_inds = np.where((in_pointcloud[:, 0] > 0) & 
                           (in_pointcloud[:, 1] > 0) & 
                           (in_pointcloud[:, 2] > 0))[0]
    points_classes = np.ones(len(in_pointcloud))
    points_classes[moving_inds] = 0
    in_pointcloud2 = in_pointcloud[moving_inds].copy()
    tp = transform3D_viewer(in_pointcloud, points_classes)
    plt.show()
        
    in_pointcloud2_transformed = tp.apply(in_pointcloud2)
    
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    ax.scatter(tp.PC[moving_inds, 0], 
               tp.PC[moving_inds, 1],
               tp.PC[moving_inds, 2], color = 'green')
    
    ax.scatter(in_pointcloud2[:, 0], 
               in_pointcloud2[:, 1], 
               in_pointcloud2[:, 2], color = 'blue')
    
    ax.scatter(in_pointcloud2_transformed[:, 0]+0.05, 
               in_pointcloud2_transformed[:, 1]+0.05, 
               in_pointcloud2_transformed[:, 2]+0.05,
               marker = 's', color = 'red')

    plt.show()    
    
    tp.figure()
    plt.show()
    print(tp.PC[moving_inds].mean(0))
    
def test_numbers_as_images():
    print('Testing function', inspect.currentframe().f_code.co_name)
    dataset_shape = (10, 10, 64, 64)
    fontsize = 10
    dataset = lognflow.plt_utils.numbers_as_images_4D(
        dataset_shape, fontsize)

    n_x, n_y, n_r, n_c = dataset_shape
    txt_width = int(np.log(np.maximum(n_x, n_y))
                    /np.log(np.maximum(n_x, n_y))) + 1
    number_text_base = '{ind_x:0{width}}, {ind_y:0{width}}'
    for ind_x, ind_y in zip([0,     n_x//3, n_x//2, n_x-1], 
                            [n_x-1, n_x//2, n_x//3, 0    ]):
        plt.figure()
        plt.imshow(dataset[ind_x, ind_y], cmap = 'gray') 
        plt.title(number_text_base.format(ind_x = ind_x, ind_y = ind_y,
                                          width = txt_width))
    plt.show()

def test_plot_gaussian_gradient():
    print('Testing function', inspect.currentframe().f_code.co_name)
    x = np.arange(0, 2, 0.1)
    mu = x**2
    std = mu**0.5

    pgg = lognflow.plt_utils.plot_gaussian_gradient()
    pgg.addPlot(x = x, mu = mu, std = std, 
                  gradient_color = (1, 0, 0), 
                  label = 'red',
                  mu_color = (0.75, 0, 0, 1),
                  mu_linewidth = 3)
    
    pgg = lognflow.plt_utils.plot_gaussian_gradient()
    pgg.addPlot(x = x, mu = mu, std = std, 
                  gradient_color = (1, 0, 0), 
                  label = 'red',
                  mu_color = (0.75, 0, 0, 1),
                  mu_linewidth = 1)
    pgg.addPlot(x = x, mu = 8 - mu, std = std, 
                gradient_color = (0, 0, 1), 
                label = 'red',
                mu_color = (0, 0, 0.75, 1),
                mu_linewidth = 1)    
    pgg.show()

def test_plt_fig_to_numpy():
    print('Testing function', inspect.currentframe().f_code.co_name)
    fig, ax = plt.subplots(111)
    ax[0].imshow(np.random.rand(100, 100))
    np_data = lognflow.plt_utils.plt_fig_to_numpy(fig)
    print(np_data.shape)
    plt.close()

def test_plt_imshow_series():
    print('Testing function', inspect.currentframe().f_code.co_name)
    data = [1 + np.random.rand(10, 100, 100),
            1 + np.random.rand(10, 10, 10)]
    mask0 = data[0][0]*0 + 1
    mask0[::2, ::2] = 0
    mask1 = data[1][0] != 0
    list_of_masks = [mask0, mask1]
    lognflow.plt_utils.plt_imshow_series(
        list_of_stacks = data, 
        list_of_masks = list_of_masks,
        list_of_titles_columns = np.arange(10),
        list_of_titles_rows = np.arange(2),
        colorbar_last_only = True,
        colorbar_labelsize = 10,
        )
    plt.show()

def test_plt_imshow_subplots():
    print('Testing function', inspect.currentframe().f_code.co_name)
    data = np.random.rand(15, 100, 100, 3)
    lognflow.plt_utils.plt_imshow_subplots(data, colorbar = False)

    data = [np.random.rand(100, 100), 
            np.random.rand(100, 150), 
            np.random.rand(50, 100)]
    lognflow.plt_utils.plt_imshow_subplots(data)

    data = np.random.rand(15, 100, 100)
    grid_locations = (np.random.rand(len(data), 2)*1000).astype('int')
    lognflow.plt_utils.plt_imshow_subplots(
        data, grid_locations = grid_locations)
    
    plt.show()

def test_plt_imshow():
    print('Testing function', inspect.currentframe().f_code.co_name)
    data = np.random.rand(100, 100) + 1j * np.random.rand(100, 100)
    plt_imshow(data, cmap = 'complex')
    plt.show()
    
def test_complex2hsv_colorbar():
    print('Testing function', inspect.currentframe().f_code.co_name)
    complex2hsv_colorbar()
    plt.show()

def test_plt_imhist():
    print('Testing function', inspect.currentframe().f_code.co_name)
    img = np.zeros((100, 100))
    indsi, indsj = np.where(img == 0)
    mask = ((indsi - 30)**2 + (indsj - 30)**2)**0.5 > 15
    mask = mask.reshape(*img.shape)
    img[mask == 0] = np.random.randn(int((mask == 0).sum()))
    img[mask == 1] = 10 + np.random.randn(int((mask == 1).sum()))
    plt_imhist(img, 
               kwargs_for_imshow = {'cmap' : 'jet'}, 
               kwargs_for_hist = {'bins': 40})
    plt.show()

def test_plt_imshow_complex():
    print('Testing function', inspect.currentframe().f_code.co_name)
    comx, comy = np.meshgrid(np.arange(-7, 8, 1), np.arange(-7, 8, 1))
    com = comx + 1j * comy
    print(comx)
    print(comy)
    img, data_abs, data_angle = complex2hsv(com)
    
    vmin = data_abs.min()
    vmax = data_abs.max()
    try:
        min_angle = data_angle[data_abs > 0].min()
    except:
        min_angle = 0
    try:
        max_angle = data_angle[data_abs > 0].max()
    except:
        max_angle = 0
    
    fig, ax = plt_imshow(img, extent=(-7, 8, -7, 8), title = 'complex2hsv',
                         colorbar = False)
    
    for i in range(0, comx.shape[0], 1):
        for j in range(0, comx.shape[1], 1):
            ax.text(j - 7+0.5, -i + 7+0.5, f'({comx[i, j]}, {comy[i, j]})', 
                    ha='center', va='center', fontsize=8, color='white')
    
    fig, ax_inset = complex2hsv_colorbar(
        (fig, ax.inset_axes([0.79, 0.03, 0.18, 0.18], transform=ax.transAxes)),
            vmin=vmin, vmax=vmax, min_angle=min_angle, max_angle=max_angle)
    ax_inset.patch.set_alpha(0)
    
    plt_imshow(np.random.rand(100, 100) + 1j * np.random.rand(100, 100),
               cmap = 'gray_real_imag')
    
    plt_imshow(np.random.rand(100, 100) + 1j * np.random.rand(100, 100),
               cmap = 'jet_real_imag')

    plt_imshow(np.random.rand(100, 100) + 1j * np.random.rand(100, 100))
    
    plt.show()

def test_plt_mark():
    print('Testing function', inspect.currentframe().f_code.co_name)
    coords = np.random.rand(1000, 2)*100
    fig, ax, markersize = plt_mark(
        coords, fig_ax=None, figsize=None,
        markersize=None, return_markersize = True)
    for cnt in range(23):
        coords = np.array([np.arange(1 + 1*cnt,1+ 1*(cnt+1)), np.zeros(1)]).T
        fig_ax = plt_mark(coords, fig_ax=(fig, ax), markersize=markersize)
    
    plt.show()

def test_plt_contours():
    print('Testing function', inspect.currentframe().f_code.co_name)
    yy, xx = np.meshgrid(np.arange(100), np.arange(50))
    Z_list = [np.exp(- ((xx-15)**2 + (yy-30)**2)**0.5 / 18), 
              np.exp(- ((xx-35)**2 + (yy-70)**2)**0.5 / 18) ]
    plt_contours(Z_list)
    plt.show()

def test_question_dialog():
    print('Testing function', inspect.currentframe().f_code.co_name)
    vec = np.random.rand(100)
    img = np.random.rand(100, 100)
    question = 'how good is it?'
    
    result = question_dialog(vec)
    print(result)
    result = question_dialog(img)
    print(result)
    result = question_dialog(question)
    print(result)

def test_stack_to_frame():
    print('Testing function', inspect.currentframe().f_code.co_name)
    data4d = np.random.rand(25, 32, 32, 3)
    img = lognflow.plt_utils.stack_to_frame(data4d, borders = np.nan)
    plt.figure()
    plt.imshow(img)
    
    data4d = np.random.rand(32, 32, 16, 16, 3)
    stack = data4d.reshape(-1, *data4d.shape[2:])
    frame = lognflow.plt_utils.stack_to_frame(stack, borders = np.nan)
    plt.figure()
    im = plt.imshow(frame)
    lognflow.plt_utils.plt_colorbar(im)
    plt.show()

def test_plt_plot():
    print('Testing function', inspect.currentframe().f_code.co_name)
    y_values_list = [[1, 2, 3], [4, 5, 6]]
    plt_plot(y_values_list, '-*')

    x_values_list = [[10, 12, 30]]
    plt_plot(y_values_list, '-*', x_values_list = x_values_list)

    y_values_list = [1, 2, 3]
    plt_plot(y_values_list, '-*', grid = dict(visible = True, which='both', linestyle='--', linewidth=2))
    
    x_values_list = [10, 12, 30]
    plt_plot(y_values_list, '-*', x_values_list = x_values_list, 
             xlim = [0, 50], ylim = [0, 10], markersize = 10, grid = True)
    
    plt.show()

def test_plt_hist2():
    print('Testing function', inspect.currentframe().f_code.co_name)
    vec1 = np.array([np.random.randn(100), np.random.randn(100) + 10])
    vec2 = np.array([np.random.randn(100), np.random.randn(100) + 10])
    data = np.array([vec1.ravel(), vec2.ravel()]).T
    
    plt_hist2(data)
    
    plt_hist2(data, use_bars = False)
    
    plt.show()

def test_plt_confusion_matrix():
    print('Testing function', inspect.currentframe().f_code.co_name)
    from sklearn.metrics import confusion_matrix
    
    n_classes = 11
    n_classes_std = 1.5
    vec1 = (np.random.rand(10000)*n_classes).astype('int')
    vec2 = (vec1 + (np.random.randn(len(vec1))*n_classes_std)).astype('int')
    vec2[vec2<0] = 0
    vec2[vec2>=n_classes] = n_classes - 1
    target_names = np.arange(n_classes)
    
    cm = confusion_matrix(vec1, vec2, normalize='all')
    plt_confusion_matrix(cm, target_names = target_names)
    plt_confusion_matrix(cm, target_names = target_names, 
                         title = 'Test_truth_accuracy_recall_precision',
                         fontsize = 8)
    plt.show()

def test_pv_volume():
    print('Testing function', inspect.currentframe().f_code.co_name)
    N = 100
    x_min, x_max = -2, 2
    y_min, y_max = -2, 2
    z_min, z_max = -2, 2
    x, y, z = np.mgrid[x_min:x_max:N*1j,
                       y_min:y_max:N*1j,
                       z_min:z_max:N*1j]
    
    sigma = 1
    gaussian = np.exp(-(x**2 + y**2 + z**2) / (2 * sigma**2))
    gaussian += np.exp(-((x-1)**2 + (y-1)**2 + z**2) / (2 * (sigma/2)**2))
    
    print(f'gaussian: {gaussian.shape}')
    plotter = pv_volume(
        gaussian, grid_size=5, grid_opacity=0.1, title="3D Gaussian with Grid")
    plotter.show()

def test_pv_surface():
    print('Testing function', inspect.currentframe().f_code.co_name)
    
    y, x = np.meshgrid(np.arange(-50, 50), np.arange(-50, 50))
    
    data1 = x**2 + y**2 + np.random.randn(100, 100)
    data2 = x**2 - y**2 + np.random.randn(100, 100)
    
    plotter = pv_surface(data1, cmap="Blues")
    
    plotter = pv_surface(data2, plotter=plotter, cmap="Reds")
    
    plotter.show()

if __name__ == '__main__':
    test_plt_imshow_subplots_comples()
    test_plt_hist2()
    test_plt_plot()
    test_plt_confusion_matrix()
    test_pv_surface()
    test_pv_volume()
    test_plt_imshow()
    test_numbers_as_images()
    test_plt_fig_to_numpy()
    test_stack_to_frame()
    test_plt_imshow_series()
    test_plt_imshow_complex()
    test_question_dialog()
    test_plt_contours()
    test_complex2hsv_colorbar()
    test_plt_mark()
    test_plot_gaussian_gradient()
    test_transform3D_viewer()
    test_plt_imshow_subplots()
    test_plt_imhist()



#!/usr/bin/env python

"""Tests for `lognflow` package."""
import pytest
from pathlib import Path
import matplotlib.pyplot as plt
import lognflow
import numpy as np
import inspect
def test_is_builtin_collection():

    # Test the function with various types
    test_list = [1, 2, 3]
    test_string = "hello"
    test_dict = {'a': 1, 'b': 2}
    test_set = {1, 2, 3}
    test_tuple = (1, 2, 3)
    test_array = np.array([1, 2, 3])
    
    assert (lognflow.is_builtin_collection(test_list))          # Expected: True
    assert not (lognflow.is_builtin_collection(test_string))    # Expected: False
    assert (lognflow.is_builtin_collection(test_dict))          # Expected: True
    assert (lognflow.is_builtin_collection(test_set))           # Expected: True
    assert (lognflow.is_builtin_collection(test_tuple))         # Expected: True
    assert not (lognflow.is_builtin_collection(test_array))     # Expected: False

def test_ssh_system():
    try:
        ssh = ssh_system(
            hostname = 'hostname', username = 'username', password = 'password')
        remote_dir = Path('/remote/folder/path')
        local_dir = Path('/local/folder/path')
        target_fname = 'intresting_file.log'
        ssh.monitor_and_move(remote_dir, local_dir, target_fname)
        ssh.close_connection()
    except:
        print('SSH test not passed maybe because you did not set the credentials.')
    
def test_printv():
    test0 = np.random.rand(10).max()
    lognflow.utils.printv(test0)
    test1 = 123
    lognflow.utils.printv(test1)
    test2 = 123.2
    lognflow.utils.printv(test2)
    test3 = np.random.rand(10000)
    lognflow.utils.printv(test3)
    test4 = [test3, 'asdf', 21]
    lognflow.utils.printv(test4)
    test5 = {'A':test3, 'B': test4}
    lognflow.utils.printv(test5)
    test55 = {'A':test5, 'C': test4}
    lognflow.utils.printv(test55)
    test6 = [test3, 'asdf', 21]*200
    lognflow.utils.printv(test6)
    test7 = 'asdf'*200000
    lognflow.utils.printv(test7)
    test8 = ['34', '123']
    lognflow.utils.printv(test8)
    test9 = np.random.rand(3, 3)
    lognflow.utils.printv(test9)

def test_save_or_load_kernel_state():
    vec = np.random.rand(100)
    vec_orig = vec.copy()
    another_variable = "Hello"
    
    current_kernel_state = lognflow.utils.save_or_load_kernel_state()
    
    vec = vec ** 2
    another_variable = "Goodbye"
    
    lognflow.utils.save_or_load_kernel_state(current_kernel_state)
    
    assert (vec == vec_orig).all()
    assert another_variable == "Hello"
    
    print("State restored successfully!")

def test_block_runner():
    from lognflow.utils import block_runner
    block_runner(Path('./test_block_runner_code.py'))

if __name__ == '__main__':
    test_is_builtin_collection()
    test_printv()
    test_block_runner()
    test_ssh_system()


#!/usr/bin/env python

"""Tests for `lognflow` package."""
import pytest
import inspect
import time
import numpy as np

from lognflow import lognflow, printprogress, select_directory

import tempfile
temp_dir = tempfile.gettempdir()

def test_log_index():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for putting index instead of time_stamp')
    logger.save('testa', 'testa', time_tag = False)
    logger.save('testb', 'testb', time_tag = True)
    logger.save('testb', 'test3', time_tag = True)
    logger.save('testc', 'test4', time_tag = 'index')
    logger.save('testc', 'test5', time_tag = 'index')
    logger.save('testc', 'test6', time_tag = 'index')
    logger.save('testd', 'test7', time_tag = 'time_and_index')
    logger.save('testd', 'test8', time_tag = 'time_and_index')
    logger.save('testd', 'test9', time_tag = 'time_and_index')
    
def test_lognflow_conflict_in_names():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for conflict in names')
    logger1 = lognflow(logger.log_dir)
    logger2 = lognflow(logger.log_dir)
    logger1(logger1.log_dir)
    logger2(logger2.log_dir)

def test_text():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir, print_text = False)
    logger('This is a test for text')    
    for _ in range(10000):
        logger(f'{_}')

    logger.text('not_main_script1.pdb',
           'This is a new log file for another script')
    logger.text('not_main_script2.test',
                    'For other log files you need to mention the log_name')
    logger.text('not_main_script3',
           'This is a new log file for another script', suffix = 'io')
    logger.text('test.to\not_main_script4.top',
                    'For other log files you need to mention the log_name')
    logger.text('not_main_script2',
           'This is a new log file for another script')
    logger.text('not_main_script2.test',
                    'For other log files you need to mention the log_name')
    logger.text('test.to\not_main_script4.top',
                    'To see if variable names are OK when passed properly')
    for _ in range(10000):
        logger(f'{_}')

def test_logger():
    print('Testing function', inspect.currentframe().f_code.co_name)
    ''' test the logger call funciton
        when lognflow object is made, you can call it.
        If it is called with a string as input, it will log that into the
        main_log text file.
        If a string is given and then something else is also provided afterwards
        the funciton save will be called with that while the first input
        string is going to be considered as the name of a variable.
    '''
    logger = lognflow(temp_dir)
    logger('This is a test for lognflow and logger call')

    a = 20
    b = np.array([34])
    c = 'asdf'
    
    logger(a)
    logger.save('a', a)
    logger.save('aa', a, suffix = 'txt', time_tag = False)
    logger(b)
    logger.save('b', b)
    logger.save('bb', b, suffix = 'txt', time_tag = False)
    logger(c)
    logger.save('test/c', c, suffix = 'txt')

def test_log_flush_period():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir, log_flush_period = 30)
    logger('This is a test for lognflow and record')    
    
    time_time = time.time()
    for _ in range(20):
        while(time.time() < time_time + 10):
            pass
        time_time = time.time()
        logger(f'Log{_}'*20)
        

    logger.text('not_main_script',
           'This is a new log file for another script')
    logger.text('not_main_script',
                    'For other log files you need to mention the log_name')

def test_record():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for lognflow and record')    

    for _ in range(1000):
        logger.record('vars/vec/v.to.txt', np.random.rand(10000))
        
def test_record_without_time_stamp():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for lognflow and record')    

    for _ in range(10):
        logger.save('vars/vec/v', np.random.rand(10000), time_tag = False)

def test_record_savefig():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for lognflow and record')    

    var = 0
    for _ in printprogress(range(1000)):
        time.sleep(0.001)
        var += np.random.randn(1)*_/50
        logger.record('vars/vec/v.to.txt', var, savefig = True)

    x = np.arange(-2, 2, 0.005)
    var = 0
    for _ in printprogress(range(len(x))):
        time.sleep(0.001)
        var = x[_]**2 + np.random.randn(1)
        logger.record('v2', var, savefig = True, plot_win_length = 10)
    
def test_get_record():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for lognflow and record')    

    var = 0
    for _ in printprogress(range(500)):
        time.sleep(0.001)
        var += np.random.rand(10000)
        logger.record('vars/vec/v.to.txt', var)
        
    time_arr, data_arr = logger.get_record('vars/vec/v.to.txt')
    logger.plot('vars/vec/v', data_arr[:, 0], '-*', x_values_list = [time_arr])
    
def test_log_animation():
    print('Testing function', inspect.currentframe().f_code.co_name)
    var1 = np.random.rand(32, 100, 100)
    logger = lognflow(temp_dir)
    logger('This is a test for log_animation')    
    logger.log_animation('var1',var1)

def test_save_matlab():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for saving MATLAB files using a dictionary')    
    data_to_save = dict(data1 = np.random.rand(3,3),
                        data2 = 'test',
                        data3 = [1, 2, 4])
    logger.save('MATLAB_test.mat', data_to_save)

def test_save():
    print('Testing function', inspect.currentframe().f_code.co_name)
    var1 = np.random.rand(100)
    
    logger = lognflow(temp_dir)
    logger('This is a test for save')    
    logger.save('var1/var1.txt', var1)
    logger.save('var1/var1.npy', var1)
    a_dict = dict({'str_var': 'This is a string',
                   'var1': var1})
    logger.save('a_dict', a_dict)
    logger.save('a_dict.txt', a_dict)
    logger.save('a_dict2', a_dict, suffix = 'txt')

def test_plot():
    print('Testing function', inspect.currentframe().f_code.co_name)
    var1 = np.random.rand(100)
    var2 = 3 + np.random.rand(100)
    var3 = 6 + np.random.rand(100)
    
    logger = lognflow(temp_dir)
    logger('This is a test for plot')    
    
    logger.plot('var1', var1)
    
    logger.plot('vars', [var1, var2, var3], '-*')
    
def test_hist():
    print('Testing function', inspect.currentframe().f_code.co_name)
    var1 = np.random.rand(10000)
    var2 = 3 + np.random.rand(10000)
    var3 = 6 + np.random.rand(10000)
    
    logger = lognflow(temp_dir)
    logger('This is a test for hist')    
    
    logger.hist(parameter_name = 'var1', 
                    parameter_value_list = var1,
                    bins = 100)
    
    logger.hist(parameter_name = 'vars', 
                    parameter_value_list = [var1, var2, var3],
                    bins = 100)
    
def test_scatter3():
    print('Testing function', inspect.currentframe().f_code.co_name)
    var1 = np.random.rand(100)
    var2 = 3 + np.random.rand(100)
    var3 = 6 + np.random.rand(100)

    var3d = np.array([var1, var2, var3])
    
    logger = lognflow(temp_dir)
    logger('This is a test for scatter3')    
    
    logger.scatter3('var3d', var3d.T)
    logger.scatter3('var3d_animation', var3d.T, make_animation=True,
                    elev_list = [45],
                    azim_list = np.arange(0, 360, 2),
                    log_animation_kwargs = {'interval' : 100})
    
def test_savefig():
    print('Testing function', inspect.currentframe().f_code.co_name)
    from lognflow.plt_utils import plt_imshow
    plt_imshow(np.random.rand(100, 100))
    logger = lognflow(temp_dir)
    logger('This is a test for savefig')    
    logger.savefig('var3d')        
    
def test_hexbin():
    print('Testing function', inspect.currentframe().f_code.co_name)
    var1 = np.random.randn(10000)
    var2 = 3 + np.random.randn(10000)

    logger = lognflow(temp_dir)
    logger('This is a test for hexbin')    
    
    logger.hexbin('hexbin', [var1, var2])    

def test_imshow():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for imshow')    
    logger.imshow('var2d', np.random.rand(100, 100))    
    logger.imshow('var2d_100_of_them', np.random.rand(25, 100, 100))

def test_surface():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for surface')    
    logger.surface('var3d', np.random.rand(100, 100))    

def test_imshow_series():
    print('Testing function', inspect.currentframe().f_code.co_name)
    """ 
        When we have 8 set of images, each is 100x100 and there are 9 of them 
        which will appear as 3x3 tile.
    """
    logger = lognflow(temp_dir)
    logger('This is a test for prepare_stack_of_images')

    stack_1 = np.random.rand(8, 9, 100, 100)
    from lognflow.plt_utils import stacks_to_frames
    
    stack_1 = stacks_to_frames(stack_1)
    stack_2 = np.random.rand(8, 100, 100)
    stack_3 = np.random.rand(8, 9, 100, 100, 3)
    stack_3 = stacks_to_frames(stack_3)
    
    list_of_stacks123 = [stack_1, stack_2, stack_3]
    
    logger.imshow_series('imshow_series_before_handling', list_of_stacks123)
    
    stack_4 = np.random.rand(1, 32, 32, 16)
    stack_4 = stacks_to_frames(stack_4)
    list_of_stacks = [stack_4]
    logger.imshow_series('just_one_series', list_of_stacks, cmap = 'cool')
    
    imgs=[]
    for _ in range(5):
        _imgs = np.random.rand(5, 100, 100)
        _imgs[:, 50, 50] = 2
        imgs.append(_imgs)
    
    logger = lognflow(temp_dir)
    logger('This is a test for imshow_series')    
    logger(f'imgs.shape: {imgs[0].shape}')

    logger.imshow_series(parameter_name = 'imshow_series\\', 
                             list_of_stacks = imgs, 
                             text_as_colorbar = True)
    
    data = [1 + np.random.rand(10, 100, 100),
            1 + np.random.rand(10, 10, 10)]
    mask0 = data[0][0]*0 + 1
    mask0[::2, ::2] = 0
    mask1 = data[1][0] != 0
    list_of_masks = [mask0, mask1]
    logger.imshow_series(parameter_name = 'imshow_series2\\',
        list_of_stacks = data, 
        list_of_masks = list_of_masks,
        list_of_titles_columns = np.arange(10),
        list_of_titles_rows = np.arange(2),
        colorbar_last_only = True,
        colorbar_labelsize = 10,
        )

def test_names_with_slashes_and_backslashes():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for test_names_with_slashes_and_backslashes')   

    _imgs = np.random.rand(10, 10)
    logger.save(r'test_param1', _imgs)
    logger.save(r'test_param2/', _imgs)
    logger.save(r'test_param3\\', _imgs)
    logger.save(r'test_param4\d', _imgs)
    logger.save(r'test_param4\d2\\', _imgs)
    logger.save(r'test_param4\d2/', _imgs)
    logger.save(r'test_param4\d2/e', _imgs)

def test_log_confusion_matrix():
    print('Testing function', inspect.currentframe().f_code.co_name)
    from sklearn.metrics import confusion_matrix
    
    n_classes = 52
    vec1 = (np.random.rand(10000)*n_classes).astype('int')
    vec2 = (np.random.rand(10000)*n_classes).astype('int')
    target_names = np.arange(n_classes)
    
    cm = confusion_matrix(vec1, vec2, normalize='all')
    logger = lognflow(temp_dir)
    logger('This is a test for log_confusion_matrix')
    logger.log_confusion_matrix('cm1', cm, target_names = target_names, 
                         title = 'Test_truth_accuracy_recall_precision')

def test_rename():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for test_rename')
    logger.rename(logger.log_dir.name + '_new_name')
    logger('This is another test for test_rename')
    
def test_save_text():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for test_save_text', flush = True)
    var = 2
    logger.save('text_log\a\t/\b/\b//\\/b', 'hello\n', suffix='txt', time_tag = False)
    logger.save('text_log\a', 'bye\n', suffix='json', time_tag = False)
    logger.save('text_log\a', var, suffix='pdb', time_tag = False)
    
def test_imshow_complex():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for test_imshow_complex', flush = True)
    
    mat = np.random.rand(100, 100) + 10 * 1j * np.random.rand(100, 100)
    
    logger(f'mat is complex? {np.iscomplexobj(mat)}')
    logger.imshow('mat', mat)
    
def test_replace_time_with_index():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for reading using logger')
    
    for _ in range(5):
        logger.save('test_param', np.array([_]))
        logger.save('testy/t', np.array([_]))
    
    flist = logger.get_flist('test_param*')
    data_in = logger.get_stack_from_files('test_param*')
    
    logger(flist)

    logger.replace_time_with_index('test_param*')
    
    flist = logger.get_flist('test_param*')
    data_out = logger.get_stack_from_files(flist = flist)
    
    logger(flist)
    
    logger(data_in)
    logger(data_out)

def test_copy_file():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for copying files')
    
    var = np.random.rand(10)
    fpath = logger.save('var', var, suffix = 'txt')
    
    logger.copy('myvar/varvar', fpath, suffix = 'pdb', 
             time_tag= True)
    
    var_check = logger.load('myvar/varvar*')
    assert str(var) == var_check
    
def test_copy_list_of_files():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for copy list of files')
    
    logger.save('test/var', np.random.rand(10), suffix = 'txt')
    logger.save('test/var', np.random.rand(10), suffix = 'fasta')
    logger.save('test/var', np.random.rand(10), suffix = 'json')
    logger.save('test/var', np.random.rand(10), suffix = 'txt')
    
    logger.copy('myvar/', 'test/var*', suffix = 'pdb', time_tag = False)
    
    for test_cnt in range(4):
        var_check1 = logger.load('test/var', file_index = test_cnt)
        var_check2 = logger.load('myvar/var', file_index = test_cnt)
        assert var_check1 == var_check2
    
def test_imshow_subplots():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for imshow_subplots')
    images = np.random.rand(20, 100, 100)
    logger.imshow_subplots('images', images, frame_shape = (4, 5))

def test_copy():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger1 = lognflow(temp_dir)
    fpath = logger1('This is a test for test_copy', flush = True)

    logger2 = lognflow(temp_dir)
    logger2.copy('some_text.txt', fpath)

def test_log_images_to_pdf():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('test log images in pdf')
    
    logger.imshow('im1', np.random.randn(30, 30))
    logger.imshow('im1', np.random.randn(20, 40))
    
    images = logger.get_stack_from_names('im1*.*')
    logger.images_to_pdf(
        'im1_all', parameter_value = images, time_tag = False)
    
def test_variables_to_pdf():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('test log variables in pdf')
    
    logger.imshow('im1', np.random.randn(30, 30))
    logger.imshow('im1', np.random.randn(20, 40))
    logger.variables_to_pdf('im1_all', 'im1*.*', time_tag = False)

def test_log_code():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('test log variables in pdf')
    
    logger.log_code(__file__)

def test_get_flist_multiple_directories():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('This is a test for test_multiple_directories_get_flist')
    
    logger.save('dir1/dir/var', np.random.rand(100))
    logger.save('dir2/dir/var', np.random.rand(100))
    logger.save('dir3/dir/var', np.random.rand(100))
    
    flist = logger.get_flist('dir*/dir/var*.npy')
    [print(_) for _ in flist]
    [print(logger.name_from_file(_)) for _ in flist]
        
def test_get_stack_from_files():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    
    logger('This is a test for get_stack_from_files')

    for _ in range(5):
        logger.save('A/img', np.random.rand(100, 100))
        logger.save('B/img', np.random.randn(100, 100))

    flist_A = logger.get_flist('A/*')
    flist_B = logger.get_flist('B/*')
    
    logger(flist_A)
    logger(flist_B)
    
    logger.replace_time_with_index('A/img')
    logger.replace_time_with_index('B/img')

    flist_A = logger.get_flist('A/*')
    flist_B = logger.get_flist('B/*')
    
    logger(flist_A)
    logger(flist_B)
    
    stack_A = logger.get_stack_from_files(flist = flist_A)
    stack_B = logger.get_stack_from_files(flist = flist_B)

    logger(f'stack_A.shape: {stack_A.shape}')
    logger(f'stack_B.shape: {stack_B.shape}')
    
    logger.imshow_series('data_samples', [stack_A, stack_B], dpi = 300)

    flist_A_AB, flist_B_AB = logger.get_common_files('A/*', 'B/*')
    logger(f'flist_A_AB: {flist_A_AB}')
    logger(f'flist_B_AB: {flist_B_AB}')
    
    if(flist_A_AB):
        
        dataset_A = logger.get_stack_from_files('A/*', flist = flist_A_AB)
        dataset_B = logger.get_stack_from_files('B/*', flist = flist_B_AB)
        
        logger.imshow_series('data_samples', 
                                 [dataset_A, dataset_B], dpi = 300)
        _ = logger._loggers_dict['log.txt'].log_size
        logger('Size of the log file in bytes is: ' \
               + f'{_}')

def test_text_to_object():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir, time_tag = False)
    test_list = ['asdf', 1243, "dd"]
    logger.save('test_list', test_list, suffix = 'txt')
    
    test_dict = {"one": "asdf", 'two': 1243, 'thre': "dd"}
    logger.save('test_dict', test_dict, suffix = 'txt')
    
    flist = logger.get_flist('*')
    print(flist)
    from lognflow.utils import text_to_collection
    for file_name_input in flist:
        print('='*60)
        print(f'file name: {file_name_input}')
        with open(file_name_input, 'r') as opened_txt:
            txt = opened_txt.read()
        print('text read from the file:')
        print(txt)
        print('- '*30)


        ext_obj = text_to_collection(txt)
        print(f'Extracted object is of type {type(ext_obj)}:')
        print(ext_obj)

def test_load_specific_fname():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('test get single specific fname')
    
    vec = np.array([1])
    logger.save('vec', vec, time_tag = False)

    vec2 = np.array([2])
    logger.save('vec2', vec2, time_tag = False)
    
    vec_out = logger.load('vec.npy')
    
    assert vec_out == vec

def test_get_stack_from_names():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('test get single specific fname')
    
    logger.imshow('im1', np.random.randn(30, 30))
    logger.imshow('im1', np.random.randn(20, 40))
    
    images = logger.get_stack_from_names('im1*')
    
    print(len(images))

def test_depricated_logviewer():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir)
    logger('testing the depricated logviewer')
    
    logger.save('test', 1)
    print(logger.logged.load('test*'))

def test_log_list_dict_read_back():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(temp_dir, time_tag = False)
    logger('testing the save for list and dict')
    try:
        import torch
        vec3 = torch.tensor([6,7,8])
    except:
        vec3 = np.array([6,7,8])
    input_list = ['test', 2, np.array([[3],[4],[5]]),
                  vec3, [9, 10, 11], {'a': 12, 'b': [13, 14, 15]}]
    logger.save('var.txt', input_list)
    var_readback = logger.load('var.txt', return_collection=True)
    
    for x, y in zip(input_list, var_readback): 
        print(f'{x} --> {y}')

def test_lognflow_here():
    print('Testing function', inspect.currentframe().f_code.co_name)
    logger = lognflow(log_dir = './')
    logger('test')

if __name__ == '__main__':
    #-----IF RUN BY PYTHON------#
    temp_dir = select_directory()
    #===========================#
    test_record_savefig()
    test_copy()
    test_replace_time_with_index()
    test_variables_to_pdf()
    test_copy_file()
    test_log_code()    
    test_log_index()
    test_copy_list_of_files()
    test_text()
    test_save_text()
    test_lognflow_conflict_in_names()
    test_rename()
    test_logger()
    test_log_flush_period()
    test_get_record()
    test_record()
    test_record_without_time_stamp()
    test_log_list_dict_read_back()
    test_save_matlab()
    test_save()
    test_names_with_slashes_and_backslashes()
    test_log_code()
    test_get_stack_from_files()
    test_get_stack_from_names()
    test_get_flist_multiple_directories()
    test_load_specific_fname()
    test_text_to_object()
    test_depricated_logviewer()
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    test_imshow()
    test_imshow_series()
    test_imshow_complex()
    test_imshow_subplots()
    test_log_images_to_pdf()
    test_log_animation()
    test_hexbin()
    test_surface()
    test_log_confusion_matrix()
    test_scatter3()
    test_plot()
    test_hist()
    test_savefig()


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec
from   matplotlib.colors import hsv_to_rgb
from   matplotlib.widgets import RangeSlider, TextBox, Button
from   mpl_toolkits.mplot3d import Axes3D
from   mpl_toolkits.axes_grid1 import make_axes_locatable
from   itertools import cycle as itertools_cycle
from   itertools import product as itertools_product

matplotlib_lines_Line2D_markers_keys_cycle = itertools_cycle([
    's', '*', 'd', 'X', 'v', '.', 'x', '|', 'D', '<','^',  '8','p',  
    '_','P','o','h', 'H', '>', '1', '2','3', '4',  '+', 'x', ])

def plt_colorbar(mappable, colorbar_aspect=3, 
                 colorbar_pad_fraction=0.05, colorbar_invisible=False, 
                 fontsize=10, tick_labels=None):
    """
    Add a colorbar to the current axis with consistent width.

    Parameters:
        mappable (AxesImage): The image to which the colorbar applies.
        colorbar_aspect (int): The aspect ratio of the colorbar width relative 
            to the axis width. Default is 3.
        colorbar_pad_fraction (float): The fraction of padding between the 
            axis and the colorbar. Default is 0.05.
        colorbar_invisible (bool): Whether to make the colorbar invisible. Default is False.
        fontsize (int): Font size for the colorbar tick labels. Default is 10.
        tick_labels (list or None): Custom labels for the colorbar ticks. 
        If None, default tick labels are used.

    Returns:
        Colorbar: The colorbar added to the axis.
    """

    ax = mappable.axes
    fig = ax.figure
    divider = make_axes_locatable(ax)
    width = ax.get_position().width / colorbar_aspect
    cax = divider.append_axes("right", size=width, pad=colorbar_pad_fraction)
    cbar = fig.colorbar(mappable, cax=cax)

    if colorbar_invisible:
        cbar.ax.set_visible(False)
    else:
        cbar.ax.tick_params(labelsize=fontsize)

        if tick_labels is not None:
            ticks = cbar.get_ticks()
            if len(tick_labels) == len(ticks):
                ticks = np.linspace(
                    mappable.get_clim()[0], mappable.get_clim()[1], len(tick_labels))
                cbar.set_ticks(ticks)
                cbar.set_ticklabels(tick_labels)

    return cbar

def complex2hsv(data_complex, vmin=None, vmax=None):
    """ complex2hsv
        Routine to visualise complex array as 2D image with color conveying
        phase information
        data_complex must be a complex 2d image
    """
    sx, sy = data_complex.shape

    data_abs = np.abs(data_complex)
    if vmin is None: vmin = data_abs.min()
    if vmax is None: vmax = data_abs.max()
    sat = (data_abs - vmin) / (vmax - vmin)
    data_angle = np.angle(data_complex) % (2 * np.pi)
    hue = data_angle / (2 * np.pi)
    a, b = np.divmod(hue, 1.0)

    H = np.zeros((sx, sy, 3))
    H[:, :, 0] = b
    H[:, :, 1] = np.ones([sx, sy])
    H[:, :, 2] = sat

    return hsv_to_rgb(H), data_abs, data_angle

def stack_to_frame(stack, frame_shape : tuple = None, borders = 0):
    """ turn a stack of images into a 2D frame of images
        This is very useful when lots of images need to be tiled
        against each other.
    
        Note: if the last dimension is 3, all images are RGB, if you don't wish that
        you have to add another dimension at the end by np.expand_dim(arr, axis = -1)
    
        :param stack: np.ndarray
                It must have the shape of either
                n_im x n_r x n_c
                n_im x n_r x  3  x  1
                n_im x n_r x n_c x  3
                
            In all cases n_im will be turned into a frame
            Remember if you have N images to put into a square, the input
            shape should be 1 x n_r x n_c x N
        :param frame_shape: tuple
            The shape of the frame to put n_rows and n_colmnss of images
            close to each other to form a rectangle of image.
        :param borders: literal or np.inf or np.nan
            When plotting images with matplotlib.pyplot.imshow, there
            needs to be a border between them. This is the value for the 
            border elements.
            
        output
        ---------
            Since we have N channels to be laid into a square, the side
            length would be ceil(N**0.5) if frame_shape is not given.
            it produces an np.array of shape n_f x n_r * f_r x n_c * f_c or
            n_f x n_r * f_r x n_c * f_c x 3 in case of an RGB input.
    """
    is_rgb = stack.shape[-1] == 3
    
    if(len(stack.shape) == 4):
        if((stack.shape[2] == 3) & (stack.shape[3] == 1)):
            stack = stack[..., 0]
    
    n_im, n_R, n_C = stack.shape[:3]
        
    if(len(stack.shape) == 4):
        assert is_rgb, 'For a stack of images with axis 3, it should be 1 or 3.'

    assert (len(stack.shape) == 3) | (len(stack.shape) == 4), \
        f'The stack you provided can have specific shapes. it is {stack.shape}'

    if(frame_shape is None):
        square_side = int(np.ceil(np.sqrt(n_im)))
        frame_n_r, frame_n_c = (square_side, square_side)
    else:
        frame_n_r, frame_n_c = frame_shape
    n_R += 2
    n_C += 2
    new_n_R = n_R * frame_n_r
    new_n_C = n_C * frame_n_c

    if is_rgb:
        frame = np.zeros((new_n_R, new_n_C, 3), dtype = stack.dtype)
    else:
        frame = np.zeros((new_n_R, new_n_C), dtype = stack.dtype)
    used_ch_cnt = 0
    if(borders is not None):
        frame += borders
    for rcnt in range(frame_n_r):
        for ccnt in range(frame_n_c):
            ch_cnt = rcnt + frame_n_c*ccnt
            if (ch_cnt<n_im):
                frame[rcnt*n_R + 1: (rcnt + 1)*n_R - 1,
                      ccnt*n_C + 1: (ccnt + 1)*n_C - 1] = \
                    stack[used_ch_cnt]
                used_ch_cnt += 1
    return frame

def stacks_to_frames(stack_list, frame_shape : tuple = None, borders = 0):
    """ turn a list of stack of images into a list of frame of images
        This is simply a list of calling stack_to_frame
        :param stack_list:
            It must have the shape of either
            n_f x n_im x n_r x n_c
            n_f x n_im x n_r x  3  x 1
            n_f x n_im x n_r x n_c x 3

    """    
    return np.array([stack_to_frame(stack, 
                                    frame_shape = frame_shape, 
                                    borders = borders) for stack in stack_list])

def plt_hist2(data, bins=30, cmap='viridis', use_bars = True,
              xlabel=None, ylabel=None, zlabel=None, title=None, 
              colorbar=True, fig_ax=None, colorbar_label=None,
              elev=None, azim=None):
    """
    Plot a 3D histogram with a colormap based on the height of the bars.

    Parameters:
    data (array-like): N x 2 array of (x, y) points.
    bins (int): Number of bins in each dimension.
    cmap (str): Name of the matplotlib colormap to use for the bars.
    xlabel (str): Label for the x-axis.
    ylabel (str): Label for the y-axis.
    zlabel (str): Label for the z-axis.
    title (str): Title of the plot.
    colorbar (bool): Whether to show a colorbar representing bar heights.
    fig_ax (tuple): Optional tuple (fig, ax) to specify the figure and axis to plot on.
    colorbar_label (str): Label for the colorbar, if shown.
    elev (float): Elevation angle in the z-plane for the 3D view.
    azim (float): Azimuthal angle in the x-y plane for the 3D view.

    Returns:
    tuple: (fig, ax) - The figure and axis objects.
    """
    
    assert data.shape[1] == 2, "Data must have shape (N, 2)"
    
    counts, x_edges, y_edges = np.histogram2d(data[:, 0], data[:, 1], bins=bins)

    x_pos, y_pos = np.meshgrid(x_edges[:-1] + 0.5 * (x_edges[1] - x_edges[0]),
                               y_edges[:-1] + 0.5 * (y_edges[1] - y_edges[0]))
    x_pos = x_pos.ravel()
    y_pos = y_pos.ravel()
    z_pos = np.zeros_like(x_pos)

    dx = dy = (x_edges[1] - x_edges[0])
    dz = counts.ravel()

    norm_dz = dz / dz.max() if dz.max() > 0 else dz

    colors = plt.cm.get_cmap(cmap)(norm_dz)

    if fig_ax is None:
        fig = plt.figure()
        if use_bars:
            ax = fig.add_subplot(111, projection='3d')
        else:
            ax = fig.add_subplot(111)
    else:
        fig, ax = fig_ax
    
    if use_bars:
        bars = ax.bar3d(x_pos, y_pos, z_pos, dx, dy, dz, 
                         color=colors, edgecolor=colors, alpha=1)
        ax.view_init(elev=elev, azim=azim)
        if colorbar:
            mappable = plt.cm.ScalarMappable(cmap=cmap)
            mappable.set_array(dz)
            cbar = plt.colorbar(mappable, ax=ax)
            if colorbar_label is not None:
                cbar.set_label(colorbar_label)
    else:
        im = ax.imshow(
            counts.T, cmap=cmap, origin='lower', aspect='auto',
            extent=[x_edges[0], x_edges[-1], y_edges[0], y_edges[-1]])
        if colorbar:
            plt_colorbar(im)

    if xlabel is not None: ax.set_xlabel(xlabel)
    if ylabel is not None: ax.set_ylabel(ylabel)
    if zlabel is not None: ax.set_zlabel(zlabel)
    if title  is not None: ax.set_title(title)
    
    return fig, ax

def _compute_tn_from_cm(cm):
    """
    Compute True Negatives (TN) directly from the confusion matrix (cm).
    True Negatives are the elements that are not in the row or column of the class.
    """
    n_classes = cm.shape[0]
    total_samples = np.sum(cm)
    tn_total = 0
    for i in range(n_classes):
        cm_without_class_i = np.delete(np.delete(cm, i, axis=0), i, axis=1)
        tn_class_i = np.sum(cm_without_class_i)
        tn_total += tn_class_i
    tn_normalized = tn_total / total_samples
    return tn_normalized

def compute_tp_tn_fp_fn(cm):
    """
    computes average tp, tn, fp and fn of a confusion matrix
    Example:
    ---------
    
    import numpy as np
    from sklearn.metrics import confusion_matrix
    from lognflow.plt_utils import plt_confusion_matrix
    
    N, n_classes = 10000, 20
    
    labels1 = (np.random.rand(N)*n_classes).astype('int')
    labels2 = (np.random.rand(N)*n_classes).astype('int')
    target_names = np.arange(n_classes)

    cm = confusion_matrix(labels1, labels2, normalize='all')

    TP, TN, FP, FN = compute_tp_tn_fp_fn(cm)
    print(TP, TN, FP, FN)
    
    """
    
    total_samples = np.sum(cm)
    TP = np.sum(np.diag(cm)) / total_samples
    FP = np.sum(np.sum(cm, axis=0) - np.diag(cm)) / total_samples
    FN = np.sum(np.sum(cm, axis=1) - np.diag(cm)) / total_samples
    TN = _compute_tn_from_cm(cm)
    
    return TP, TN, FP, FN

def calculate_contrasting_color(value, cmap):
    """Calculate a contrasting color for a given value in the colormap."""
    r, g, b, _ = cmap(value)  # Get the RGBA values
    # Calculate luminance using the sRGB luminance formula
    luminance = 0.2126 * r + 0.7152 * g + 0.0722 * b
    # Return white for dark colors and black for light colors
    return 'white' if luminance < 0.5 else 'black'

def plt_confusion_matrix(cm, 
        target_names=None, title=None, cmap=None,
        figsize=None, fontsize = None):
    """
    This function plots a confusion matrix and returns the figure and axis.
    Parameters:
    - cm: Confusion matrix
    - target_names: List of target names (default: None)
    - title: have any of the following options you like to see in the title
        title = YOUR_TITLE + '_truth' + '_accuracy' + '_recall' + '_precision' + '_f1_score' + '_specificity' + '_mcc'
        YOUR_TITLE could be nothing like ''
        eg. if I want to see the truth table and the f1_score I code:
            title = 'My truth table' + '_truth_f1_score'
        
    - cmap: Colormap (default: None)
    - figsize: Size of the figure (default: None)
    Returns:
    - fig: Figure object
    - ax: Axis object
    
    
    Example:
    ---------
    
    import numpy as np
    from sklearn.metrics import confusion_matrix
    from lognflow.plt_utils import plt_confusion_matrix
    
    N, n_classes = 10000, 20
    
    vec1 = (np.random.rand(N)*n_classes).astype('int')
    vec2 = (np.random.rand(N)*n_classes).astype('int')
    target_names = np.arange(n_classes)

    cm = confusion_matrix(vec1, vec2, normalize='all')

    plt_confusion_matrix(cm, target_names = target_names)
    plt.show()
    
    """
    TP, TN, FP, FN = compute_tp_tn_fp_fn(cm)

    figsize_was_None = False
    if figsize is None:
        figsize = np.ceil(cm.shape[0]**0.5)
        figsize_was_None = True
    if figsize_was_None:
        figsize = np.maximum(figsize, 5)
        figsize = (figsize + 1, figsize)
    if target_names is None:
        target_names = [chr(x + 65) for x in range(cm.shape[0])]

    if cmap is None:
        cmap = plt.get_cmap('Blues')

    fig, ax = plt.subplots(figsize=figsize)
    
    if fontsize is None:
        fontsize = 6*(np.minimum(*figsize) / cm.shape[0])**0.5

    im = ax.imshow(cm, interpolation='nearest', cmap=cmap)
    
    tick_marks = np.arange(len(target_names))
    ax.set_xticks(tick_marks)
    ax.set_xticklabels(target_names, rotation=45, fontsize = fontsize)
    ax.set_yticks(tick_marks)
    ax.set_yticklabels(target_names, fontsize = fontsize)
    for i, j in itertools_product(range(cm.shape[0]), range(cm.shape[1])):
        clr = calculate_contrasting_color(cm[i, j] / cm.max(), cmap)
        ax.text(j, i, f"{cm[i, j]:2.02f}", horizontalalignment="center", color=clr,
                fontsize = fontsize)

    for i in range(cm.shape[0]):
        rect = plt.Rectangle((i - 0.5, i - 0.5), 1, 1, fill=False, 
                             edgecolor='darkred', lw=1, linestyle='--')
        ax.add_patch(rect)

    ax.set_ylabel('True label', fontsize = fontsize)
    ax.set_xlabel('Predicted label', fontsize = fontsize)
    if title is None:
        title_str = f'Average TP={TP:.02f}, TN={TN:.02f}, FP={FP:.02f}, FN={FN:.02f}'
    else:
        title_str = ''
        if '_truth' in title.lower():
            title_str += f'Average TP={TP:.02f}, TN={TN:.02f}, FP={FP:.02f}, FN={FN:.02f}\n'
            title = title.replace('_truth', '')
        if '_accuracy' in title.lower():
            accuracy = (TP + TN) / (TP + TN + FP + FN)
            title_str += f'Accuracy={accuracy:.04f}'
            title = title.replace('_accuracy', '')
        if '_recall' in title.lower():
            recall = TP / (TP + FN) if (TP + FN) > 0 else 0
            title_str += f', Recall={recall:.04f}'
            title = title.replace('_recall', '')
        if '_precision' in title.lower():
            precision = TP / (TP + FP) if (TP + FP) > 0 else 0
            title_str += f', Precision={precision:.04f}'
            title = title.replace('_precision', '')
        if '_f1_score' in title.lower():
            f1_score = (2 * precision * recall) / (precision + recall) if (precision + recall) > 0 else 0
            title_str += f', F1-Score={f1_score:.04f}'
            title = title.replace('_f1_score', '')
        if '_specificity' in title.lower():
            specificity = TN / (TN + FP) if (TN + FP) > 0 else 0
            title_str += f', Specificity={specificity:.04f}'
            title = title.replace('_specificity', '')
        if '_mcc' in title.lower():
            numerator = (TP * TN) - (FP * FN)
            denominator = ((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN)) ** 0.5
            mcc = numerator / denominator if denominator > 0 else 0
            title_str += f', MCC={mcc:.04f}'
            title = title.replace('_mcc', '')
            
        if len(title) > 0:
            if title[-1] != '\n':
                title += '\n'
        title_str = title + title_str
        
    ax.set_title(title_str, fontsize = fontsize)
    plt_colorbar(im, colorbar_invisible=False, colorbar_aspect = fontsize**0.5 * 4, 
                 fontsize = fontsize, tick_labels = target_names)
    return fig, ax

def complex2hsv_colorbar(
        fig_and_ax=None, vmin=0, vmax=1, 
        min_angle=0, max_angle=0, 
        fontsize=8, angle_threshold=np.pi / 18):
    
    xx, yy = np.meshgrid(
        np.linspace(-1, 1, 1000),
        np.linspace(-1, 1, 1000))
    conv, sat, _ = complex2hsv(xx + 1j * yy, vmax=1)

    # Set outside the circle to transparent
    mask = (xx ** 2 + yy ** 2) > 1
    conv_rgba = np.zeros((conv.shape[0], conv.shape[1], 4))
    conv_rgba[..., :3] = conv
    conv_rgba[..., 3] = 1.0  # Set alpha to 1 for everything
    conv_rgba[mask, 3] = 0  # Set alpha to 0 outside the circle
    conv_rgba[conv_rgba < 0] = 0
    conv_rgba[conv_rgba > 1] = 1
    conv_rgba = conv_rgba[::-1, :]
    if fig_and_ax is None:
        fig, ax = plt.subplots()
    else:
        try:
            fig, ax = fig_and_ax
        except Exception as e:
            print('fig_and_ax should be a two-tuple of (fig, ax). Use:')
            print('>>> fig, ax = plt.subplots()')
            raise e

    im = ax.imshow(conv_rgba, interpolation='nearest')  # Flip the image vertically
    ax.axis('off')

    diff = np.abs(max_angle - min_angle)
    # Draw lines at min and max angles if they are not too close
    if np.minimum(diff, 2 * np.pi - diff) > angle_threshold:
        for angle in [min_angle, max_angle]:
            x_end = 500 + np.cos(angle) * 500
            y_end = 500 - np.sin(angle) * 500
            ax.plot([500, x_end], [500, y_end], '--', color='gray')

    # Add text annotations for min and max values
    if int(vmin*100)/100 > 0:   #because we are going to show .2f
        ax.text(500, 500, f'{vmin:.2f}', 
                ha='center', va='center', fontsize=fontsize, color='white')

    # Calculate position for max value text and invert color for readability
    angle = 45 * np.pi / 180  # 45 degrees in radians
    x_max = int(np.cos(angle) * 500 + 300)
    y_max = int(np.sin(angle) * 500 - 200)

    bck_color = conv_rgba[y_max, x_max, :3]
    text_color = 1 - bck_color  # Invert color

    ax.text(x_max, y_max, f'{vmax:.2f}',
            ha='center', va='center', fontsize=fontsize, color=text_color)

    return fig, ax

def plt_violinplot(
        dataset:list, positions, facecolor = None, edgecolor = None, 
        alpha = 0.5, label = None, fig_and_ax : tuple = None, 
        title = None, plt_violinplot_kwargs = {}):
    
    if(fig_and_ax is None):
        fig, ax = plt.subplots(1)
    else:
        fig, ax = fig_and_ax
    violin_parts = ax.violinplot(dataset, positions, **plt_violinplot_kwargs)
    for partname in ('cbars','cmins','cmaxes','cmeans','cmedians','bodies'):
        vp = violin_parts.get(partname, [])
        if partname == 'bodies':
            for vp_body in vp:
                vp_body.set_facecolor(facecolor)
                vp_body.set_edgecolor(edgecolor)
                vp_body.set_alpha(alpha)
        else:
            if isinstance(vp, list):
                for v in vp:
                    v.set_edgecolor(facecolor)
            else:
                vp.set_edgecolor(facecolor)

    if title is not None:
        title = str(title)
        fig.suptitle(title)
        fig.canvas.manager.window.setWindowTitle(title)

    return fig, ax

class plt_imhist:
    def __init__(self, in_image, figsize=(12, 6), title=None, bins=None,
                 kwargs_for_imshow={}, kwargs_for_hist={}):
        if bins is not None:
            if not (bins in kwargs_for_hist):
                kwargs_for_hist['bins'] = bins
        
        try:
            in_image = in_image.detach().cpu().numpy()
            print('plt_imhist warning: '
                  'image converted from torch to numpy for plt!')
        except: pass

        # Adjust figsize to provide more space if needed
        self.fig, axs = plt.subplots(
            1, 2, figsize=figsize,
            gridspec_kw={'width_ratios': [5, 1], 'wspace': 0.1})
        self.fig.subplots_adjust(left=0.05, right=0.85, bottom=0.1, top=0.9)
        
        self.fig_ax = self.fig, axs[0]
        
        # Display the image
        self.im = axs[0].imshow(in_image, **kwargs_for_imshow)
        if title is not None:
            title = str(title)
            axs[0].set_title(title)
        axs[0].axis('off')
        
        cm = self.im.get_cmap()
        
        # Histogram
        im_image_ravel = in_image.ravel().copy()
        im_image_ravel = im_image_ravel[np.isnan(im_image_ravel) == False]
        im_image_ravel = im_image_ravel[np.isinf(im_image_ravel) == False]
        if len(im_image_ravel) == 0:
            im_image_ravel = in_image.ravel().copy()
        n, bins = np.histogram(im_image_ravel, **kwargs_for_hist)
        bin_centres = 0.5 * (bins[:-1] + bins[1:])
        axs[1].barh(
            bin_centres, n, height=(bins[1]-bins[0]),
            color=cm((bin_centres - bin_centres.min()) /
                         (bin_centres.max() - bin_centres.min())))
        axs[1].invert_xaxis()
        
        axs[1].yaxis.set_visible(True)
        axs[1].xaxis.set_visible(False)
        
        # Create textbox axes
        upper_text_ax = self.fig.add_axes([0.88, 0.85, 0.05, 0.05])
        lower_text_ax = self.fig.add_axes([0.88, 0.1, 0.05, 0.05])
        
        self.upper_text_box = TextBox(
            upper_text_ax, 'Max', initial=f'{in_image.max():.6f}')
        self.lower_text_box = TextBox(
            lower_text_ax, 'Min', initial=f'{in_image.min():.6f}')
        
        # Calculate the position for the slider
        slider_top = 0.85 - 0.02  # Bottom of the upper text box
        slider_bottom = 0.1 + 0.07  # Top of the lower text box
        slider_height = slider_top - slider_bottom  # Height between the two text boxes
        
        # Create slider axes on the right side of the histogram
        slider_ax = self.fig.add_axes(
            [0.895, slider_bottom, 0.02, slider_height], 
            facecolor='lightgoldenrodyellow')
        self.slider = RangeSlider(
            slider_ax, '', in_image.min(), in_image.max(),
            valinit=[in_image.min(), in_image.max()], orientation='vertical')
        self.slider.label.set_visible(False)
        self.slider.valtext.set_visible(False)
        
        self.lower_limit_line = axs[1].axhline(
            self.slider.val[0], color='k', linestyle='--')
        self.upper_limit_line = axs[1].axhline(
            self.slider.val[1], color='k', linestyle='--')
        
        # Initial text annotations for vmin and vmax
        self.vmin_text = axs[1].text(
            0.5, self.slider.val[0], f'{self.slider.val[0]:.6f}',
            transform=axs[1].get_yaxis_transform(), 
            ha='right', va='bottom', color='k')
        self.vmax_text = axs[1].text(
            0.5, self.slider.val[1], f'{self.slider.val[1]:.6f}',
            transform=axs[1].get_yaxis_transform(),
            ha='right', va='top', color='k')
        
        self.slider.on_changed(self.update)
        self.lower_text_box.on_submit(self.update_from_text)
        self.upper_text_box.on_submit(self.update_from_text)
    
    def update(self, val):
        self.im.set_clim(val[0], val[1])
        self.lower_limit_line.set_ydata([val[0], val[0]])
        self.upper_limit_line.set_ydata([val[1], val[1]])
        
        # Update text annotations to reflect the new vmin and vmax
        self.vmin_text.set_position((0.5, val[0]))
        self.vmin_text.set_text(f'{val[0]:.6f}')
        self.vmax_text.set_position((0.5, val[1]))
        self.vmax_text.set_text(f'{val[1]:.6f}')
        
        # Update text boxes to reflect the new values
        self.lower_text_box.set_val(f'{val[0]:.6f}')
        self.upper_text_box.set_val(f'{val[1]:.6f}')
        
        self.fig.canvas.draw_idle()
    
    def update_from_text(self, text):
        try:
            lower_val = float(self.lower_text_box.text)
            upper_val = float(self.upper_text_box.text)
            if lower_val < upper_val:
                self.slider.set_val([lower_val, upper_val])
        except ValueError:
            pass

def _listify_1d_list(list_of_obj):
    if list_of_obj is not None:
        if len(list_of_obj) > 1:
            it_is_1d = True
            for _ in list_of_obj:
                try:
                    if len(_) > 1:
                        it_is_1d = False
                except: pass
            if it_is_1d:
                list_of_obj = [np.array(list_of_obj).squeeze()]
    return list_of_obj

def plt_plot(y_values_list, *plt_plot_args, x_values_list = None, 
             fig_ax = None, title = None, **kwargs):
    """
        Plots multiple sets of y-values against x-values using Matplotlib, 
        with options to customize the plot.
    
        Parameters
        ----------
        y_values_list : list or iterable
            A list or iterable of y-values to plot. If a single 
            iterable is provided, it will be treated as 
            one dataset. If a list of iterables is provided, 
            each will be plotted as a separate line.
        
        x_values_list : list or iterable
            The x-values for the plot. This can either be a list 
            of the same length as `y_values_list`, 
            or a single iterable that will be reused for all y-values. 
            If `None`, y-values will be plotted 
            against their index.
        
        *plt_plot_args : tuple
            Additional positional arguments passed to the Matplotlib 
            `plot` function (e.g., line style, 
            marker type).
        
        fig_ax : tuple (figure, axes), optional
            A tuple containing a Matplotlib `figure` and `axes` object. 
            If `None`, a new figure and axes 
            will be created.
        
        title : str, optional
            The title of the plot. If `None`, no title will be set.
        
        **kwargs : dict
            Additional keyword arguments passed to the Matplotlib 
            `plot` function (e.g., `color`, `linewidth`).
        
        Returns
        -------
        tuple
            A tuple containing the Matplotlib `figure` and `axes`
             objects used for the plot.
        
        Raises
        ------
        ValueError
            If the length of `x_values_list` does not match the length of 
            `y_values_list` or if it is not 1.
        
        Notes
        -----
        - If `fig_ax` is provided, the plot will be added to the 
        given axes. Otherwise, a new figure and axes 
          will be created.
        - If `x_values_list` is `None`, y-values will be plotted against their index.
        - The function can handle multiple y-value datasets, plotting 
        each as a separate line.
        
        Example
        -------
        >>> y_values_list = [[1, 2, 3], [4, 5, 6]]
        >>> x_values_list = [1, 2, 3]
        >>> fig, ax = plt_plot(y_values_list, x_values_list)
        >>> plt.show()
    """
    y_values_list = _listify_1d_list(y_values_list)
    x_values_list = _listify_1d_list(x_values_list)
    
    if x_values_list is not None:
        assert ( (len(x_values_list) >= len(y_values_list)) | \
                 (len(x_values_list) == 1) ), \
                f'lognflow plt_plot: x_values_list has length {len(x_values_list)},'\
                ' should have length of 1 or the same as parameters list: '\
                f'{len(y_values_list)}.'
    
    if fig_ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        fig, ax = fig_ax
    
    if 'xlim' in kwargs:
        xlim = kwargs['xlim']
        kwargs.pop('xlim')
        ax.set_xlim(xlim)

    if 'ylim' in kwargs:
        ylim = kwargs['ylim']
        kwargs.pop('ylim')
        ax.set_ylim(ylim)
    
    use_grid = None
    if 'grid' in kwargs:
        use_grid = kwargs['grid']
        kwargs.pop('grid')    
    
    for list_cnt, y_values in enumerate(y_values_list):
        if(x_values_list is None):
            ax.plot(y_values, *plt_plot_args, **kwargs)
        else:
            if(len(x_values_list) == len(y_values)):
                x_values = x_values_list[list_cnt]
            else:
                x_values = x_values_list[0]
            ax.plot(x_values, y_values, *plt_plot_args, **kwargs)

    if use_grid:
        if isinstance(use_grid, dict):
            ax.grid(**use_grid)
        else:
            ax.grid(use_grid)

    if title is not None:
        title = str(title)
        ax.set_title(title)
        
    return (fig, ax)

def plt_imshow(img, 
               fig_ax = None,
               colorbar = True, 
               remove_axis_ticks = False, 
               title = None, 
               cmap = None,
               angle_cmap = None,
               portrait = None,
               aspect = 'equal',
               **kwargs):
    """
    Display an image or a complex-valued image using matplotlib's imshow.

    This function can handle real images and complex-valued data, allowing for
    visualization of magnitude and phase. The function provides options for 
    displaying a colorbar, removing axis ticks, and setting titles. If the input 
    image is complex, it will be represented in either RGB or separate real and 
    imaginary components.

    Parameters:
    ----------
    img : array_like
        The image data to be displayed. This can be a 2D array for real images or 
        a 2D complex array for complex-valued data.
        
    fig_ax : tuple, optional
        A tuple containing a figure and an axis to plot on. If None, a new figure 
        and axis will be created.
        
    colorbar : bool, optional
        Whether to display a colorbar alongside the image. Default is True.
        
    remove_axis_ticks : bool, optional
        Whether to remove ticks from the axes. Default is False.
        
    title : str, optional
        A title to be displayed above the figure. Default is None.
        
    cmap : str, optional
        The colormap to be used for displaying the image. Default is None.
        to get real and imag part separately for a xomplex image, use 
        cmap = 'complex_real_imag', if you don't provide the cmap, it will show
        the abs and angle part of the image separately.
        
    angle_cmap : str, optional
        The colormap to be used for displaying the angle of complex numbers. 
        Default is twilight_shifted. 
        
    portrait : bool, optional
        If True, the figure will be set up in portrait mode. If None, the function 
        will automatically determine the orientation based on the window dimensions.
    
    aspect : str, optional
        by default I set the aspect ratio to equal
    
    **kwargs : keyword arguments
        Additional keyword arguments passed to `imshow`, such as `vmin`, `vmax`, 
        etc.

    Returns: 2-tuple
    -------
    fig : matplotlib.figure.Figure
        The figure object containing the displayed image(s).
        
    ax : matplotlib.axes.Axes or list of Axes
        The axes object(s) containing the displayed image(s). If the image is 
        complex and displayed as two separate plots, a list of axes will be returned.
    """
    
    vmin = kwargs['vmin'] if 'vmin' in kwargs else None
    vmax = kwargs['vmax'] if 'vmax' in kwargs else None
    
    try:
        img = img.detach().cpu().numpy()
        print('plt_imshow warning: image converted from torch to numpy for plt!')
    except: pass
    
    if(not np.iscomplexobj(img)):
        if fig_ax is None:
            fig, ax = plt.subplots()
        else:
            fig, ax = fig_ax
        im = ax.imshow(img, cmap = cmap, **kwargs)
        if(colorbar):
            plt_colorbar(im)
        if(remove_axis_ticks):
            plt.setp(ax, xticks=[], yticks=[])
        if aspect is not None:
            ax.set_aspect(aspect)
    else:
        if (cmap == 'complex'):
                
            complex_image, data_abs, data_angle = complex2hsv(
                img, vmin = vmin, vmax = vmax)
        
            if vmin is None: vmin = data_abs.min()
            if vmax is None: vmax = data_abs.max()
            
            try:
                min_angle = data_angle[data_abs > 0].min()
            except:
                min_angle = 0
            try:
                max_angle = data_angle[data_abs > 0].max()
            except:
                max_angle = 0
        
            if fig_ax is None:
                fig, ax = plt.subplots()
            else:
                fig, ax = fig_ax
            im = ax.imshow(complex_image)
            if(remove_axis_ticks):
                plt.setp(ax, xticks=[], yticks=[])

            if(colorbar):
                fig, ax_inset = complex2hsv_colorbar(
                    (fig, ax.inset_axes([0.79, 0.03, 0.18, 0.18], 
                                        transform=ax.transAxes)),
                    vmin=vmin, vmax=vmax, min_angle=min_angle, max_angle=max_angle)
                ax_inset.patch.set_alpha(0)
            if aspect is not None:
                ax.set_aspect(aspect)
        else:
            
            if fig_ax is None:
                fig = plt.figure()
            else:
                fig, _ = fig_ax
            
            try:
                window = plt.get_current_fig_manager().window
                if (window.height() > window.width()) & (portrait is None):
                    portrait = True
            except: pass
            if portrait:
                ax = [fig.add_subplot(2, 1, 1), fig.add_subplot(2, 1, 2)]
            else:
                ax = [fig.add_subplot(1, 2, 1), fig.add_subplot(1, 2, 2)]
            
            complex_real_imag = False
            if cmap is not None:
                if 'real_imag' in cmap:
                    complex_real_imag = True
            if complex_real_imag:
                cmap = cmap.split('real_imag')[0]
                if len(cmap) == 0: cmap = None
                else: cmap = cmap[:-1]
                if angle_cmap is None:
                    angle_cmap = cmap
                im = ax[0].imshow(np.real(img), cmap = cmap, **kwargs)
                if(colorbar):
                    plt_colorbar(im)
                ax[0].set_title('real')
                im = ax[1].imshow(np.imag(img), cmap = angle_cmap, **kwargs)
                if(colorbar):
                    plt_colorbar(im)
                ax[1].set_title('imag')
            else:
                im = ax[0].imshow(np.abs(img), cmap = cmap, **kwargs)
                if(colorbar):
                    plt_colorbar(im)
                ax[0].set_title('abs')    
                if angle_cmap is None:
                    angle_cmap = 'twilight_shifted'
                im = ax[1].imshow(np.angle(img), cmap = angle_cmap, **kwargs)
                if(colorbar):
                    plt_colorbar(im)
                ax[1].set_title('angle')
                            
            if(remove_axis_ticks):
                plt.setp(ax[0], xticks=[], yticks=[])
                ax[0].xaxis.set_ticks_position('none')
                ax[0].yaxis.set_ticks_position('none')
                plt.setp(ax[1], xticks=[], yticks=[])
                ax[1].xaxis.set_ticks_position('none')
                ax[1].yaxis.set_ticks_position('none')
            
            if aspect is not None:
                ax[0].set_aspect(aspect)
                ax[1].set_aspect(aspect)
    
    if title is not None:
        title = str(title)
        fig.suptitle(title)
        fig.canvas.manager.window.setWindowTitle(title)
        
    return fig, ax

def plt_hist(vectors_list, fig_ax = None,
             bins = 10, alpha = 0.5, normalize = False, 
             labels_list = None, **kwargs):
    
    if fig_ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        fig, ax = fig_ax
    
    for vec_cnt, vec in enumerate(vectors_list):
        bins_, edges = np.histogram(vec, bins)
        if normalize:
            bins_ = bins_ / bins_.max()
        ax.bar(edges[:-1], bins_, 
                width =np.diff(edges).mean(), alpha=alpha)
        if labels_list is None:
            ax.plot(edges[:-1], bins_, **kwargs)
        else:
            assert len(labels_list) == len(vectors_list)
            ax.plot(edges[:-1], bins_, 
                     label = f'{labels_list[vec_cnt]}', **kwargs)
    return fig, ax

def plt_hist_subplots(arrays, bins = 10, frame_shape=None, alpha=0.7, 
                      kwargs_bar = {} , kwargs_plot = {}):
    """
    Function to plot histograms of multiple arrays in subplots.
    
    Parameters:
    - arrays (list of np.ndarray): List of arrays to plot histograms for.
    - bins (int or sequence of scalars): Bin specification for the histograms.
    - frame_shape (tuple or None): If provided, specify the number of rows and columns for the subplots.
    - alpha (float): The transparency of the bars.
    - color (str): The color of the bars.
    - **kwargs: Additional arguments passed to the plot.
    """
    
    N = len(arrays)
    if frame_shape is None:
        cols = int(np.ceil(np.sqrt(N)))
        rows = int(np.ceil(N / cols))
    else:
        rows, cols = frame_shape
        N = min(N, rows * cols)
        
    fig, axes = plt.subplots(rows, cols, figsize=(cols * 5, rows * 5))
    axes = axes.flatten()
    
    for cnt, data in enumerate(arrays):
        ax = axes[cnt]
        bins_data, edges = np.histogram(data, bins=bins)
        
        ax.bar(edges[:-1], bins_data, width=np.diff(edges).mean(), 
               alpha=alpha, **kwargs_bar)
        
        ax.plot(edges[:-1], bins_data, **kwargs_plot)
        
        ax.set_title(f"Histogram {cnt + 1}")

    # Hide unused axes if any
    for ax in axes[N:]:
        ax.set_visible(False)

    plt.tight_layout()
    

def plt_scatter3(
        data_N_by_3, fig_ax = None, title = None, 
        elev_list = [20, 70], azim_list = np.arange(0, 360, 20),
        make_animation = False, **kwargs):
    assert (len(data_N_by_3.shape)==2) & (data_N_by_3.shape[1] == 3), \
        'The first argument must be N x 3'
    if fig_ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
    else:
        fig, ax = fig_ax
    ax.scatter(data_N_by_3[:, 0], 
               data_N_by_3[:, 1], 
               data_N_by_3[:, 2], **kwargs)
    
    if title is not None:
        title = str(title)
        ax.set_title(title)
        fig.canvas.manager.window.setWindowTitle(title)

    try: elev_list = [int(elev_list)]
    except: pass
    try: azim_list = [int(azim_list)]
    except: pass

    if make_animation:
        stack = []
        for elev in elev_list:
            for azim in azim_list:
                ax.view_init(elev=elev, azim=azim)
                img = plt_fig_to_numpy_3ch(fig)
                stack.append(img)
        return fig, ax, stack
    else:
        elev = None if elev_list is None else elev_list[0]
        azim = None if azim_list is None else azim_list[0]
        if (elev is not None) | (azim is not None):
            ax.view_init(elev=elev, azim=azim)
        return fig, ax

def plt_surface(stack, fig_ax = None, **kwargs):
    n_r, n_c = stack.shape

    if fig_ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
    else:
        fig, ax = fig_ax

    X, Y = np.meshgrid(np.arange(n_r, dtype='int'), 
                       np.arange(n_c, dtype='int'))
    ax.plot_surface(X, Y, stack, **kwargs)
    return fig, ax

def plt_fig_to_numpy_3ch(fig):
    """Convert a matplotlib figure to a numpy 2D array (RGB)."""
    fig.canvas.draw()
    w, h = fig.canvas.get_width_height()
    buf = np.frombuffer(fig.canvas.tostring_argb(), dtype=np.uint8)
    buf.shape = (h, w, 4)  # Shape should be (height, width, 4) for RGBA
    buf = np.copy(buf)  # Ensure we have a copy, not a view
    return buf

def plt_fig_to_numpy(fig):
    """ from https://www.icare.univ-lille.fr/how-to-
                    convert-a-matplotlib-figure-to-a-numpy-array-or-a-pil-image/
    """
    fig.canvas.draw()
    w,h = fig.canvas.get_width_height()
    buf = np.frombuffer(fig.canvas.tostring_argb(), dtype=np.ubyte)
    buf.shape = (w, h, 4)
    return buf.sum(2)

def numbers_as_images_3D(data3D_shape: tuple,
                         fontsize: int, 
                         text_loc: tuple = None,
                         verbose: bool = True):
    """ Numbers3D
    This function generates a 4D dataset of images with shape
    (n_x, n_r, n_c) where in each image the value "x" is written as a text
    that fills the image. As such, later when working with such a dataset you can
    look at the image and know which index it had before you use it.
    
    Follow this recipe to make good images:
    
    1- set n_x to 10, Set the desired n_r, n_c and width. 
    2- find fontsize that is the largest and still fits
    3- Increase n_x to desired size.
    
    You can provide a logs_root, log_dir or simply select a directory to save the
    output 3D array.
    
    """
    n_x, n_r, n_c = data3D_shape
    
    if text_loc is None:
        text_loc = (n_r//2 - fontsize, n_c//2 - fontsize)
    
    dataset = np.zeros(data3D_shape)    
    txt_width = int(np.log(n_x)/np.log(n_x)) + 1
    number_text_base = '{ind_x:0{width}}}'
    if(verbose):
        from lognflow import printprogress
        pBar = printprogress(n_x)
    for ind_x in range(n_x):
        mat = np.ones((n_r, n_c))
        number_text = number_text_base.format(ind_x = ind_x, 
                                              width = txt_width)
        fig = plt.figure(figsize = (n_rr, n_cc), dpi = n_rc)
        ax = fig.add_subplot(111)
        ax.imshow(mat, cmap = 'gray', vmin = 0, vmax = 1)
        ax.text(text_loc[0], text_loc[1],
                number_text, fontsize = fontsize)
        ax.axis('off')
        buf = plt_fig_to_numpy(fig)
        plt.close()
        dataset[ind_x] = buf.copy()
        if(verbose):
            pBar()
    return dataset

def numbers_as_images_4D(data4D_shape: tuple,
                         fontsize: int, 
                         text_loc: tuple = None,
                         verbose: bool = True):
    """ Numbers4D
    This function generates a 4D dataset of images with shape
    (n_x, n_y, n_r, n_c) where in each image the value "x, y" is written as a text
    that fills the image. As such, later when working with such a dataset you can
    look at the image and know which index it had before you use it.
    
    Follow this recipe to make good images:
    
    1- set n_x, n_y to 10, Set the desired n_r, n_c and width. 
    2- try fontsize that is the largest
    3- Increase n_x and n_y to desired size.
    
    You can provide a logs_root, log_dir or simply select a directory to save the
    output 4D array.
    
    :param text__loc:
        text_loc should be a tuple of the location of bottom left corner of the
        text in the image.
    
    """
    n_x, n_y, n_r, n_c = data4D_shape

    if text_loc is None:
        text_loc = (n_r//2 - fontsize, n_c//2 - fontsize)
    
    dataset = np.zeros((n_x, n_y, n_r, n_c))    
    txt_width = int(np.log(np.maximum(n_x, n_y))
                    / np.log(np.maximum(n_x, n_y))) + 1
    number_text_base = '{ind_x:0{width}}, {ind_y:0{width}}'
    if(verbose):
        from lognflow import printprogress
        pBar = printprogress(n_x * n_y)
    for ind_x in range(n_x):
        for ind_y in range(n_y):
            mat = np.ones((n_r, n_c))
            number_text = number_text_base.format(
                ind_x = ind_x, ind_y = ind_y, width = txt_width)
            n_rc = np.minimum(n_r, n_c)
            n_rr = n_r / n_rc
            n_cc = n_c / n_rc
            fig = plt.figure(figsize = (n_rr, n_cc), dpi = n_rc)
            ax = fig.add_subplot(111)
            ax.imshow(mat, cmap = 'gray', vmin = 0, vmax = 1)
            ax.text(text_loc[0], text_loc[1], number_text, fontsize = fontsize)
            ax.axis('off')
            buf = plt_fig_to_numpy(fig)
            plt.close()
            dataset[ind_x, ind_y] = buf.copy()
            if(verbose):
                pBar()
    return dataset

class plot_gaussian_gradient:
    """ Orignally developed for RobustGaussinFittingLibrary
    Plot curves by showing their average, and standard deviatoin
    by shading the area around the average according to a Gaussian that
    reduces the alpha as it gets away from the average.
    You need to init() the object then add() plots and then show() it.
    refer to the tests.py
    """
    def __init__(self, xlabel = None, ylabel = None, num_bars = 100, 
                       title = None, xmin = None, xmax = None, 
                       ymin = None, ymax = None, fontsize = 14):
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.title = title
        self.num_bars = num_bars
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        LWidth = 1
        font = {
                'weight' : 'bold',
                'size'   : fontsize}
        plt.rc('font', **font)
        params = {'legend.fontsize': 'x-large',
                 'axes.labelsize': 'x-large',
                 'axes.titlesize':'x-large',
                 'xtick.labelsize':'x-large',
                 'ytick.labelsize':'x-large'}
        plt.rcParams.update(params)
        plt.figure(figsize=(8, 6), dpi=50)
        self.ax1 = plt.subplot(111)
    
    def addPlot(self, x, mu, std, gradient_color, label, 
                snr = 3.0, mu_color = None, general_alpha = 1,
                mu_linewidth = 1):

        for idx in range(self.num_bars-1):
            y1 = ((self.num_bars-idx)*mu + idx*(mu + snr*std))/self.num_bars
            y2 = y1 + snr*std/self.num_bars
            
            prob = np.exp(-(snr*idx/self.num_bars)**2/2)
            plt.fill_between(
                x, y1, y2, 
                color = (gradient_color + (prob*general_alpha,)), 
                edgecolor=(gradient_color + (0,)))

            y1 = ((self.num_bars-idx)*mu + idx*(mu - snr*std))/self.num_bars
            y2 = y1 - snr*std/self.num_bars
            
            plt.fill_between(
                x, y1, y2, 
                color = (gradient_color + (prob*general_alpha,)), 
                edgecolor=(gradient_color + (0,)))
        if(mu_color is None):
            mu_color = gradient_color
        plt.plot(x, mu, linewidth = mu_linewidth, color = mu_color, 
                 label = label)
        
    def show(self, show_legend = True):
        if(self.xmin is not None) & (self.xmax is not None):
            plt.xlim([self.xmin, self.xmax])
        if(self.ymin is not None) & (self.ymax is not None):
            plt.ylim([self.ymin, self.ymax])
        if(self.xlabel is not None):
            plt.xlabel(self.xlabel, weight='bold')
        if(self.ylabel is not None):
            plt.ylabel(self.ylabel, weight='bold')
        if(self.title is not None):
            plt.title(self.title)
        if(show_legend):
            plt.legend()
        plt.grid()
        
        plt.show()
        
    def __call__(self, *args, **kwargs):
        self.addPlot(*args, **kwargs)

def plt_imshow_series(list_of_stacks, 
                      list_of_masks = None,
                      figsize = None,
                      text_as_colorbar = False,
                      colorbar = False,
                      cmap = 'viridis',
                      list_of_titles_columns = None,
                      list_of_titles_rows = None,
                      fontsize = None,
                      vmin = None,
                      vmax = None,
                      title = None,
                      colorbar_last_only = True,
                      colorbar_fraction = 0.046,
                      colorbar_pad = 0.04,
                      colorbar_labelsize = 1,
                      grid_width_space = 0.0,
                      remove_axis_ticks = True,
                      aspect = 'equal',
                      **kwargs,
                      ):
    
    """
    Displays a grid of image series for comparison with optional customization for annotations, colorbars, and formatting.
    
    Parameters:
        list_of_stacks (list): 
            A list of 3D or 4D arrays, each representing a stack of images. 
            All stacks must have the same number of images.
            
        list_of_masks (list, optional): 
            A list of masks corresponding to the stacks. Each mask should have the same shape 
            as the images in its respective stack. If provided, masked areas will be ignored 
            when calculating statistics. Defaults to None.
            
        figsize (tuple, optional): 
            The overall size of the figure in inches. If None, it is determined based on 
            the number of stacks and images. Defaults to None.
            
        text_as_colorbar (bool, optional): 
            If True, displays the maximum, mean, and minimum values of each image as text 
            in place of a colorbar. Defaults to False.
            
        colorbar (bool, optional): 
            If True, displays a colorbar for each subplot. Defaults to False.
            
        cmap (str, optional): 
            The colormap to use for displaying the images. Defaults to 'viridis'.
            
        list_of_titles_columns (list, optional): 
            Titles for each column in the grid. Must have a length equal to the number 
            of images in each stack. Defaults to None.
            
        list_of_titles_rows (list, optional): 
            Titles for each row in the grid. Must have a length equal to the number of stacks. 
            Defaults to None.
            
        fontsize (int, optional): 
            Font size for the text annotations. If None, it is determined based on the figure size. 
            Defaults to None.
            
        vmin (float, optional): 
            The minimum value for image normalization. If None, it is automatically calculated 
            from the image data. Defaults to None.
            
        vmax (float, optional): 
            The maximum value for image normalization. If None, it is automatically calculated 
            from the image data. Defaults to None.
            
        title (str, optional): 
            The title for the entire figure. Defaults to None.
            
        colorbar_last_only (bool, optional): 
            If True, displays a colorbar only for the last column. Defaults to False.
            
        colorbar_fraction (float, optional): 
            Fraction of the original axis allocated for the colorbar. Defaults to 0.046.
            
        colorbar_pad (float, optional): 
            Padding between the image and colorbar. Defaults to 0.04.
            
        colorbar_labelsize (int, optional): 
            Label size for the colorbar. Defaults to 1.
            
        grid_width_space (float, optional): 
            Horizontal spacing between grid columns. Defaults to 0.0.
            
        remove_axis_ticks (bool, optional): 
            If True, removes axis ticks from all subplots. Defaults to True.
            
        aspect (str, optional): 
            Aspect ratio of the displayed images. Defaults to 'equal'.
            
        **kwargs: 
            Additional keyword arguments to pass to the `imshow` function.
    
    Returns:
        tuple:
            - fig (matplotlib.figure.Figure): The created figure.
            - None: Placeholder for potential additional return values.
            
    Raises:
        AssertionError: 
            If the input lists do not meet the expected shapes or lengths.
    """
    
    if colorbar_last_only:
        colorbar = False
    
    n_stacks = len(list_of_stacks)
    if(list_of_masks is not None):
        assert len(list_of_masks) == n_stacks, \
            f'the number of masks, {len(list_of_masks)} and ' \
            + f'stacks {n_stacks} should be the same'
     
    n_imgs = len(list_of_stacks[0])
    for ind, stack in enumerate(list_of_stacks):
        assert len(stack) == n_imgs, \
            'All members of the given list should have same number of images.' \
            f' while the stack indexed as {ind} has length {len(stack)}.'
        assert (len(stack.shape) == 3) | (len(stack.shape) == 4), \
            f'The shape of the stack {ind} must have length 3 or 4, it has '\
            f'shape of {stack.shape}. Perhaps you wanted to have only '\
             'one set of images. If thats the case, put that single '\
             'image in a list.'

    if (list_of_titles_columns is not None):
        assert len(list_of_titles_columns) == n_imgs, \
            f'len(list_of_titles_columns): {len(list_of_titles_columns)}, ' \
            + f'should be len(list_of_stacks[0]): {n_imgs}'

    if (list_of_titles_rows is not None):
        assert len(list_of_titles_rows) == n_stacks, \
            f'len(list_of_titles_rows): {len(list_of_titles_rows)}, ' \
            + f'should be len(list_of_stacks): {n_stacks}'
            
    if figsize is None:
        if(colorbar):
            figsize = (n_imgs* 2, n_stacks)
        else:
            figsize = (n_imgs, n_stacks)

    if fontsize is None:
        fontsize = int(max(figsize)/10)
        if fontsize > 8: fontsize = 8
    
    fig = plt.figure(figsize = figsize)
    if colorbar_last_only:
        gs1 = matplotlib.gridspec.GridSpec(n_stacks, n_imgs + 1)
    else:
        gs1 = matplotlib.gridspec.GridSpec(n_stacks, n_imgs)
    if grid_width_space:
        gs1.update(wspace=grid_width_space, hspace=0)
    
    for img_cnt in range(n_imgs):
        for stack_cnt in range(n_stacks):
            ax = plt.subplot(gs1[stack_cnt, img_cnt])
            
            data_canvas = list_of_stacks[stack_cnt][img_cnt].copy()
            if(list_of_masks is not None):
                mask = list_of_masks[stack_cnt]
                if(mask is not None):
                    if(data_canvas.shape == mask.shape):
                        data_canvas[mask==0] = 0
                        data_canvas_stat = data_canvas[mask>0]
            else:
                data_canvas_stat = data_canvas.copy()
            data_canvas_stat = data_canvas_stat[
                np.isnan(data_canvas_stat) == 0]
            data_canvas_stat = data_canvas_stat[
                np.isinf(data_canvas_stat) == 0]
            if vmin is None:
                vmin = data_canvas_stat.min()
            if vmax is None:
                vmax = data_canvas_stat.max()

            im = ax.imshow(data_canvas, 
                           cmap = cmap, vmin = vmin, vmax = vmax, **kwargs)
            if(remove_axis_ticks):
                plt.setp(ax, xticks=[], yticks=[])
            
            if aspect is not None:
                ax.set_aspect(aspect)
            
            if colorbar | colorbar_last_only:
                if colorbar_last_only:
                    colorbar_invisible = img_cnt != n_imgs - 1
                else:
                    colorbar_invisible = 0
                plt_colorbar(im, colorbar_invisible = colorbar_invisible)
                            
            if(text_as_colorbar):
                ax.text(data_canvas.shape[0]*0,
                         data_canvas.shape[1]*0.05,
                         f'{data_canvas.max():.6f}', 
                         color = 'yellow',
                         fontsize = fontsize)
                ax.text(data_canvas.shape[0]*0,
                         data_canvas.shape[1]*0.5, 
                         f'{data_canvas.mean():.6f}', 
                         color = 'yellow',
                         fontsize = fontsize)
                ax.text(data_canvas.shape[0]*0,
                         data_canvas.shape[1]*0.95, 
                         f'{data_canvas.min():.6f}', 
                         color = 'yellow',
                         fontsize = fontsize)
            
            if (list_of_titles_rows is not None):
                if img_cnt == 0:
                    ax.set_ylabel(list_of_titles_rows[stack_cnt])
            if (list_of_titles_columns is not None):
                if stack_cnt == 0:
                    ax.set_title(list_of_titles_columns[img_cnt])
            
    if title is not None:
        title = str(title)
        fig.suptitle(title)
        fig.canvas.manager.window.setWindowTitle(title)
    return fig, None

def plt_imshow_subplots(
        images, grid_locations=None, frame_shape = None, title = None,
        titles=[], cmaps=[], colorbar=True, margin = 0.025,
        inter_image_margin = 0.01,
        colorbar_aspect=2, colorbar_pad_fraction=0.05,
        figsize=None, remove_axis_ticks=True, **kwargs):
    """
    Plots a list of 2D images at specified 2D grid_locations with titles 
    and colormaps.
    
    Parameters:
    images (list of 2D arrays): List of 2D images to plot.
    grid_locations (list of tuples or None): List of subplot grid_locations 
        in (rows, cols, index) format or None to generate a grid.
    titles (list of str): List of titles for each image.
    cmaps (list of str): List of colormaps for each image.
    colorbar (bool): Whether to add a colorbar beside each image. 
        Default is True.
    colorbar_aspect (int): Aspect ratio for the colorbars. Default is 2.
    colorbar_pad_fraction (float): Padding fraction for the colorbars. 
        Default is 0.05.
    figsize (tuple): Size of the figure.
    remove_axis_ticks (bool): Whether to remove axis ticks. Default is True.
    """
    try:
        dims = images.shape
        if len(dims) == 2:
            dims = [dims]
    except: pass
    
    if colorbar:
        margin = np.maximum(margin, 0.4)
        inter_image_margin = np.maximum(margin, 0.4)
    
    N = len(images)
    # Determine the maximum image size
    max_width = max(img.shape[1] for img in images)
    max_height = max(img.shape[0] for img in images)
    
    if grid_locations is None:
        if frame_shape is None:
            cols = int(np.ceil(np.sqrt(N)))
            rows = int(np.ceil(N / cols))
        else:
            rows, cols = frame_shape
            N = np.minimum(N, rows * cols)
        
        # Generate grid locations with dynamic spacing
        spacing = max(max_width, max_height) * (1 + inter_image_margin)
        grid_locations = np.array([[col * spacing, 1 - row * spacing] for row in range(rows) for col in range(cols)])
        grid_locations = grid_locations[:N]  # Trim to number of images
            
    lefts = grid_locations[:, 0]
    bottoms = grid_locations[:, 1]
    rights = lefts + np.array([img.shape[1] for img in images])
    tops = bottoms + np.array([img.shape[0] for img in images])
    min_left = lefts.min() - margin * max_width
    min_bottom = bottoms.min() - margin * max_height
    max_right = rights.max() + margin * max_width
    max_top = tops.max() + margin * max_height
    lefts = (lefts - min_left) / (max_right - min_left)
    bottoms = (bottoms - min_bottom) / (max_top - min_bottom)
    rights = (rights - min_left) / (max_right - min_left)
    tops = (tops - min_bottom) / (max_top - min_bottom)

    fig = plt.figure(figsize = figsize)
    for cnt in range(N):
        gs = matplotlib.gridspec.GridSpec(1, 1, left=lefts[cnt], right=rights[cnt], 
                                          top=tops[cnt], bottom=bottoms[cnt])
        ax = fig.add_subplot(gs[0])
        image = images[cnt]
        
        try:
            image = image.detach().cpu().numpy()
        except: pass
        
        if image is not None:
            if np.iscomplexobj(image):
                complex_image, data_abs, data_angle = complex2hsv(image)
                try: min_angle = data_angle[data_abs > 0].min()
                except: min_angle = 0
                try: max_angle = data_angle[data_abs > 0].max()
                except: max_angle = 0
                cax = ax.imshow(complex_image)
                
                if(colorbar):
                    vmin = kwargs.get('vmin', np.abs(image).min())
                    vmax = kwargs.get('vmax', np.abs(image).max())
                    
                    fig, ax_inset = complex2hsv_colorbar(
                        (fig, ax.inset_axes([0.79, 0.03, 0.18, 0.18], 
                                            transform=ax.transAxes)),
                        vmin=vmin, vmax=vmax, min_angle=min_angle, max_angle=max_angle)
                    ax_inset.patch.set_alpha(0)
                
            else:
                cmap = None
                if cmaps:
                    cmap = cmaps[i]
                else:
                    if 'cmap' in kwargs:
                        cmap = kwargs['cmap']
                if cmap:
                    cax = ax.imshow(image, cmap=cmap, **kwargs)
                else:
                    cax = ax.imshow(image, **kwargs)
                if colorbar:
                    plt_colorbar(cax, colorbar_aspect=colorbar_aspect,
                                 colorbar_pad_fraction=colorbar_pad_fraction)
            
            try: ax.set_title(titles[cnt])
            except: pass
            
            if remove_axis_ticks:
                ax.axis('off')    
            
    if title is not None:
        title = str(title)
        fig.suptitle(title)
        fig.canvas.manager.window.setWindowTitle(title)
    
    plt.subplots_adjust(left=0, right=1, top=1, bottom=0)
    plt.margins(margin)
    return fig, ax

class transform3D_viewer:
    """
    A 3D viewer for point cloud transformations using matplotlib.

    Attributes:
        in_pointcloud (numpy.ndarray): The input point cloud.
        pt_cls (numpy.ndarray): there must be a class for each point.
            class 0 is movable others will only have different colors
    """
    def __init__(self, in_pointcloud, pt_cls = None):
        error_msg = 'input point cloud must be Nx3, where N >= 3'
        assert len(in_pointcloud.shape) == 2, error_msg
        assert in_pointcloud.shape[0] >= 3, error_msg
        assert in_pointcloud.shape[1] == 3, error_msg
        self.PC = in_pointcloud
    
        if pt_cls is None:
            pt_cls = np.zeros(len(in_pointcloud), dtype='int')
        self.pt_cls = pt_cls
        self.moving_inds = np.where(self.pt_cls == 0)[0]
        assert len(self.moving_inds) > 3, \
            'at least 3 data points must have class 0'
        self.params = {}
        self.figure()
        self.textboxevalues = np.array([
            float(self.params["Tx_text_box"].text),
            float(self.params["Ty_text_box"].text),
            float(self.params["Tz_text_box"].text),
            float(self.params["Sx_text_box"].text),
            float(self.params["Sy_text_box"].text),
            float(self.params["Sz_text_box"].text),
            float(self.params["Rx_text_box"].text),
            float(self.params["Ry_text_box"].text),
            float(self.params["Rz_text_box"].text)])

    def figure(self):
        from functools import partial

        self.Theta_init, self.Vt_init = self.get_Theta(self.PC[self.moving_inds])
        
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111, projection='3d')
        self.fig.subplots_adjust(left=0.05, right=0.5, bottom=0.1, top=0.9)
        
        # Create step size text boxes
        self.create_text_box("T_step", 0.75, 0.88, 1.0, self.update_steps)
        # Create transformation widgets
        self.create_text_box("Tx", 0.75, 0.81, self.Theta_init["Tx"], self.update_from_text)
        self.create_text_box("Ty", 0.75, 0.74, self.Theta_init["Ty"], self.update_from_text)
        self.create_text_box("Tz", 0.75, 0.67, self.Theta_init["Tz"], self.update_from_text)

        self.create_buttons("Tx", 0.70, 0.81, partial(self.update_value, "Tx", "T_step", -1), partial(self.update_value, "Tx", "T_step", 1))
        self.create_buttons("Ty", 0.70, 0.74, partial(self.update_value, "Ty", "T_step", -1), partial(self.update_value, "Ty", "T_step", 1))
        self.create_buttons("Tz", 0.70, 0.67, partial(self.update_value, "Tz", "T_step", -1), partial(self.update_value, "Tz", "T_step", 1))
        
        self.create_text_box("S_step", 0.75, 0.59, 0.1, self.update_steps)
        self.create_text_box("Sx", 0.75, 0.52, self.Theta_init["Sx"], self.update_from_text)
        self.create_text_box("Sy", 0.75, 0.45, self.Theta_init["Sy"], self.update_from_text)
        self.create_text_box("Sz", 0.75, 0.38, self.Theta_init["Sz"], self.update_from_text)
        
        self.create_buttons("Sx", 0.70, 0.52, partial(self.update_value, "Sx", "S_step", -1), partial(self.update_value, "Sx", "S_step", 1))
        self.create_buttons("Sy", 0.70, 0.45, partial(self.update_value, "Sy", "S_step", -1), partial(self.update_value, "Sy", "S_step", 1))
        self.create_buttons("Sz", 0.70, 0.38, partial(self.update_value, "Sz", "S_step", -1), partial(self.update_value, "Sz", "S_step", 1))
        
        self.create_text_box("R_step", 0.75, 0.3, 5.0, self.update_steps)
        self.create_text_box("Rx", 0.75, 0.23, self.Theta_init["Rx"], self.update_from_text)
        self.create_text_box("Ry", 0.75, 0.18, self.Theta_init["Ry"], self.update_from_text)
        self.create_text_box("Rz", 0.75, 0.13, self.Theta_init["Rz"], self.update_from_text)
        
        self.create_buttons("Rx", 0.70, 0.23, partial(self.update_value, "Rx", "R_step", -1), partial(self.update_value, "Rx", "R_step", 1))
        self.create_buttons("Ry", 0.70, 0.18, partial(self.update_value, "Ry", "R_step", -1), partial(self.update_value, "Ry", "R_step", 1))
        self.create_buttons("Rz", 0.70, 0.13, partial(self.update_value, "Rz", "R_step", -1), partial(self.update_value, "Rz", "R_step", 1))

        self.draw()
        
    def draw(self):
        # Display the point cloud
        self.ax.cla()
        for cls_cnt in np.unique(self.pt_cls):
            self.ax.scatter(self.PC[self.pt_cls == cls_cnt, 0],
                            self.PC[self.pt_cls == cls_cnt, 1],
                            self.PC[self.pt_cls == cls_cnt, 2], 
                            label=f'cls_{cls_cnt}')
        cls_values = np.unique(self.pt_cls)
        if len(cls_values) > 1:
            for cls_cnt in cls_values[:-1] :
                self.ax.plot([self.PC[self.pt_cls == cls_cnt, 0][-1], self.PC[self.pt_cls == cls_cnt + 1, 0][0]],
                             [self.PC[self.pt_cls == cls_cnt, 1][-1], self.PC[self.pt_cls == cls_cnt + 1, 1][0]],
                             [self.PC[self.pt_cls == cls_cnt, 2][-1], self.PC[self.pt_cls == cls_cnt + 1, 2][0]], 
                             color = 'black', linewidth = 2)
    
        # Calculate the bounding box for the moving_inds using SVD
        points = self.PC[self.moving_inds]
        mean = points.mean(axis=0)
        centered_points = points - mean
        U, S, Vt = np.linalg.svd(centered_points)
    
        # Project points onto principal axes
        projections = centered_points @ Vt.T
    
        # Get the min and max along each principal axis
        min_proj = projections.min(axis=0)
        max_proj = projections.max(axis=0)
    
        # Define the bounding box corners in the projected space
        bbox_proj = np.array([[min_proj[0], min_proj[1], min_proj[2]],
                              [max_proj[0], min_proj[1], min_proj[2]],
                              [max_proj[0], max_proj[1], min_proj[2]],
                              [min_proj[0], max_proj[1], min_proj[2]],
                              [min_proj[0], min_proj[1], max_proj[2]],
                              [max_proj[0], min_proj[1], max_proj[2]],
                              [max_proj[0], max_proj[1], max_proj[2]],
                              [min_proj[0], max_proj[1], max_proj[2]]])
    
        # Rotate bounding box corners back to the original coordinate system
        bbox = bbox_proj @ Vt + mean
    
        # Draw bounding box lines
        edges = [(0, 1), (1, 2), (2, 3), (3, 0), # Bottom square
                 (4, 5), (5, 6), (6, 7), (7, 4), # Top square
                 (0, 4), (1, 5), (2, 6), (3, 7)] # Vertical lines
    
        for edge in edges:
            self.ax.plot3D(*zip(bbox[edge[0]], bbox[edge[1]]), '--', color='blue')
    
        self.fig.canvas.draw()

    def get_Theta(self, PC):
        from   scipy.spatial.transform import Rotation as scipy_rotation
        Theta = {}
        mean_vec = PC.mean(0)
        Theta["Tx"], Theta["Ty"], Theta["Tz"] = mean_vec
        PC_moving_centered = PC - mean_vec
        U, S_vec, Vt = np.linalg.svd(PC_moving_centered.T)
        Theta["Sx"], Theta["Sy"], Theta["Sz"] = S_vec
        r = scipy_rotation.from_matrix(U)
        Theta["Rx"], Theta["Ry"], Theta["Rz"] = r.as_euler('xyz', degrees=True)
        return Theta, Vt[:3]
    
    def apply(self, PC):
        from   scipy.spatial.transform import Rotation as scipy_rotation
        Theta_in, Vt_in = self.get_Theta(PC)
        Theta, _ = self.get_Theta(self.PC[self.moving_inds])
        
        translation = np.array(
            [Theta_in['Tx'] + Theta["Tx"] - self.Theta_init['Tx'],
             Theta_in['Ty'] + Theta["Ty"] - self.Theta_init['Ty'],
             Theta_in['Tz'] + Theta["Tz"] - self.Theta_init['Tz']])
        new_S = np.diag(
            [Theta_in["Sx"] * Theta["Sx"] / self.Theta_init["Sx"],
             Theta_in["Sy"] * Theta["Sy"] / self.Theta_init["Sy"],
             Theta_in["Sz"] * Theta["Sz"] / self.Theta_init["Sz"]])
        r = scipy_rotation.from_euler('xyz',
            np.array([Theta_in["Rx"] + Theta["Rx"] - self.Theta_init["Rx"],
                      Theta_in["Ry"] + Theta["Ry"] - self.Theta_init["Ry"],
                      Theta_in["Rz"] + Theta["Rz"] - self.Theta_init["Rz"]]),
            degrees=True)
        new_U = r.as_matrix()
        PC_transformed = (new_U @ new_S @ Vt_in).T + translation
        return PC_transformed
        
    def create_text_box(self, label, x, y, initial_val, on_submit):
        text_ax = self.fig.add_axes([x, y, 0.13, 0.05])
        text_box = TextBox(text_ax, label + '         ', 
                           initial=f'{initial_val:.6f}')
        text_box.on_submit(on_submit)
        self.params[f"{label}_text_box"] = text_box
    
    def create_buttons(self, label, x, y, on_click_minus, on_click_plus):
        minus_ax = self.fig.add_axes([x, y, 0.04, 0.05])
        plus_ax = self.fig.add_axes([x + 0.19, y, 0.04, 0.05])
        minus_button = Button(minus_ax, '-')
        plus_button = Button(plus_ax, '+')
        minus_button.on_clicked(on_click_minus)
        plus_button.on_clicked(on_click_plus)
        self.params[f"{label}_minus_button"] = minus_button
        self.params[f"{label}_plus_button"] = plus_button
    
    def update_steps(self, text):
        try:
            self.params["T_step"] = float(self.params["T_step_text_box"].text)
            self.params["S_step"] = float(self.params["S_step_text_box"].text)
            self.params["R_step"] = float(self.params["R_step_text_box"].text)
        except ValueError:
            pass

    def update_from_text(self, text):
        from   scipy.spatial.transform import Rotation as scipy_rotation
        try: # Read new transformation values
            self.textboxevalues = np.array([
                float(self.params["Tx_text_box"].text),
                float(self.params["Ty_text_box"].text),
                float(self.params["Tz_text_box"].text),
                float(self.params["Sx_text_box"].text),
                float(self.params["Sy_text_box"].text),
                float(self.params["Sz_text_box"].text),
                float(self.params["Rx_text_box"].text),
                float(self.params["Ry_text_box"].text),
                float(self.params["Rz_text_box"].text)])
        except ValueError:
            pass
            
        translation = self.textboxevalues[:3].copy()
        new_S = np.diag(self.textboxevalues[3:6].copy())
        r = scipy_rotation.from_euler(
            'xyz',self.textboxevalues[6:].copy(), degrees=True)
        new_U = r.as_matrix()
        PC_transformed = (new_U @ new_S @ self.Vt_init).T + translation
        # Update the movable part of the point cloud
        self.PC[self.moving_inds] = PC_transformed
        
        self.draw()

    def update_value(self, label, step_label, direction, event):
        current_val = float(self.params[f"{label}_text_box"].text)
        step_size = float(self.params[f"{step_label}_text_box"].text)
        new_val = current_val + direction * step_size
        self.params[f"{label}_text_box"].set_val(f"{new_val:.6f}")

class _questdiag:
    def __init__(self, question, buttons, figsize, question_hratio):
        
        assert isinstance(buttons, dict), \
            ('buttons arg must be a dictionary of texts appearing on '
             'the buttons values to be returned.')
        
        self.buttons = buttons
        self.result = None

        # Calculate the number of rows and columns for the buttons
        N = len(self.buttons)
        n_rows = int(np.ceil(N ** 0.5))  # Number of rows for buttons
        n_cols = int(np.ceil(N / n_rows))  # Number of columns for buttons
        
        if N == 1: n_rows, n_cols = 1, 1
        if N == 2: n_rows, n_cols = 1, 2
        if N == 3: n_rows, n_cols = 1, 3
        if N == 6: n_rows, n_cols = 2, 3
        
        if question_hratio is None:
            if isinstance(question, np.ndarray):
                question_hratio = 10
            else:
                question_hratio = 1
        
        if figsize is None:
            if isinstance(question, np.ndarray):
                figsize = (7.5, 7.5)
            else:
                figsize = (5, 2.5)
            
        # Create the figure and GridSpec layout
        fig = plt.figure(figsize=figsize)
        gs = matplotlib.gridspec.GridSpec(n_rows + 2, n_cols, 
                      figure=fig, 
                      height_ratios=[question_hratio] + [1] * (n_rows + 1))  
        # First row (3x height) for the question, remaining rows for buttons
        
        # Top section for the question (span the entire width)
        ax_question = fig.add_subplot(gs[0, :])
        
        # Handle different types of questions
        if isinstance(question, np.ndarray):
            if len(question.shape) == 1:
                ax_question.plot(question)
            elif len(question.shape) == 2:
                plt_imshow(question, fig_ax=(fig, ax_question))
            plt.axis('on')  # Keep axis on for plots and images
        else:
            ax_question.text(0.5, 0.5, str(question), 
                             ha='center', va='center', fontsize=12)
            ax_question.set_axis_off()  # No axis for text questions

        # Create buttons and place them on the grid
        button_objects = []
        for i, (label, val) in enumerate(self.buttons.items()):
            row = 2 + i // n_cols
            col = i % n_cols
            button_ax = fig.add_subplot(gs[row, col])
            button = Button(button_ax, label)
            button.on_clicked(self.button_click)
            button_objects.append(button)
    
        plt.show()
    
    def button_click(self, event):
        ind = event.inaxes.texts[0].get_text()  # Get text of the clicked button
        self.result = self.buttons[ind]  # Return the corresponding output
        plt.close()  # Close the plot after a button is clicked

def question_dialog(
    question = 'Yes/No/Cancel?',
    buttons={'Yes': True, 'No': False, 'Cancel': None},
    figsize = None, 
    question_hratio = None):
    """ Question dialog
    Creates a dialog with a question displayed at the top and a grid of buttons below it.
    
    The function supports displaying questions as text, 1D numpy arrays (as line plots),
    or 2D numpy arrays (as images). It displays buttons beneath the question, allowing the
    user to select one of the provided options. The buttons are organized into a grid
    layout based on the number of buttons provided. When a button is clicked, the function 
    returns the corresponding value associated with the button in the `buttons` dictionary.

    Parameters
    ----------
    question : str, np.ndarray, optional
        The question to be presented. It can be a string, a 1D numpy array (plotted as a 
        line), or a 2D numpy array (displayed as an image). Default is 'Yes/No/Cancel?'.
    
    buttons : dict, optional
        A dictionary where the keys are the text labels that will appear on the buttons, 
        and the values are the corresponding values to return when the button is clicked.
        Default is {'Yes': True, 'No': False, 'Cancel': None}.
        
    figsize : tuple, optional
        A tuple specifying the size of the figure (width, height) in inches. Default is (6, 2).

    question_hratio: int, optional
        If you are sending an image as a question, you can set the height ratio to
        buttons here, we suggest 4
    Returns
    -------
    result : any
        The value associated with the button clicked by the user. If 'Yes' is clicked, 
        returns `True`; if 'No', returns `False`; and if 'Cancel', returns `None`.
    """
    return _questdiag(question, buttons, figsize, question_hratio).result

def plt_mark(
        coords, fig_ax=None, figsize=(2, 2),
        marker=None, markersize = None, return_markersize = False):
    """
    Plots a grid of dots with a dynamic figure size to avoid overlap.
    
    Parameters:
    - coords: numpy array of shape (N, 2), where each row is [x, y] coordinates
    - fig_ax: 2-tuple of (fig, ax) or None; if None, a new figure and axis are created
    - figsize: tuple of two floats, figure size in inches (width, height)
    - marker: str, marker style (e.g., 'x', 'o', '.', etc.); if None, use the next marker in the cycle
    - marker_sizer: float, the marker size
    
    Returns:
    - 2-tuple of (fig, ax), and the markersize used for plotting
    """
    if fig_ax is None:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot()
        if figsize is None:
            figsize = fig.get_size_inches()
    else:
        fig, ax = fig_ax
    
    if markersize is None:
        markersize = 12 * min(figsize[0], figsize[1])/ len(coords)
        markersize = np.maximum(markersize, 1)

    if marker is None:
        marker = next(matplotlib_lines_Line2D_markers_keys_cycle)
    
    ax.plot(coords[:, 0], coords[:, 1], 
            marker=marker, markersize=markersize, linestyle='')

    if return_markersize:
        return fig, ax, markersize
    else:
        return fig, ax
    
def plt_contours(
        Z_list, X_Y = None, fig_ax = None, levels = 10, colors_list = None, 
        linestyles_list = None, linewidth = 0.5, fontsize = 3, title = None):
    """
    Plot contours of multiple surfaces overlaid on the same plot.
    
    Parameters:
    - Z_list: List of 2D arrays representing the surface heights at each 
              grid point.
    - X_Y: tuple where the (X, Y) describe the meshgrid over which Z is defined
    - fig_ax: Tuple (fig, ax) where fig is the figure and ax is the axes.
              If None, creates a new figure and axes.
    - levels: Number of contour levels for all surfaces.
    - colors_list: List of colors for the contours of each surface. 
                   If None, defaults to a colormap.
    - linestyles_list: List of line styles for the contours of each surface. 
                       If None, defaults to a pattern.
    - title: Optional title for the plot.
    """
    
    # Create figure and axes if not provided
    if fig_ax is None:
        fig, ax = plt.subplots()
    else:
        fig, ax = fig_ax
    
    # Default colors and linestyles if not provided
    if colors_list is None:
        colors_list = plt.cm.jet(np.linspace(0, 1, len(Z_list)))
    if linestyles_list is None:
        linestyles_list = ['dashed', 'solid'] * (len(Z_list) // 2 + 1)
    
    # Plot contours for each surface in Z_list
    for i, Z in enumerate(Z_list):
        if X_Y is None:
            Y, X = np.meshgrid(np.arange(Z.shape[1]), np.arange(Z.shape[0]))
        else:
            X, Y = X_Y
        color = colors_list[i % len(colors_list)]
        linestyle = linestyles_list[i % len(linestyles_list)]
        contour = ax.contour(X, Y, Z, levels=levels, colors=[color],
                             linestyles=linestyle, linewidths = linewidth)
        
        # Add labels to contours
        ax.clabel(contour, inline=True, fontsize=fontsize, fmt='%.2f')
        
    ax.set_aspect('equal')

    if title is not None:
        title = str(title)
        ax.set_title(title)
        fig.canvas.manager.window.setWindowTitle(title)
    
    return fig, ax

def pv_volume(volume, volume_xyz = None, 
                   grid_size=None, grid_opacity=0.5, show_grid=True, 
                   title="volume Visualization", show_ticks = True):
    import pyvista as pv

    if volume_xyz is None:
        x, y, z = np.mgrid[:volume.shape[0],
                           :volume.shape[1],
                           :volume.shape[2]]
    else:
        x, y, z = volume_xyz
    
    x_min, x_max, y_min, y_max, z_min, z_max = (x.min(), x.max(), y.min(), y.max(), z.min(), z.max()) 
    normed_values = (volume - volume.min()) / (volume.max() - volume.min())

    points = np.column_stack((x.ravel(), y.ravel(), z.ravel()))
    point_cloud = pv.PolyData(points)
    point_cloud["volume"] = normed_values.ravel()

    plotter = pv.Plotter()
    plotter.add_title(title)

    plotter.add_points(point_cloud, scalars="volume", cmap="viridis", opacity="sigmoid")
    if show_ticks:
        plotter.show_bounds(
            grid='front',
            location='outer',
            all_edges=True,
        )

    return plotter

def pv_surface(data_2d, plotter=None, show_edges=True, cmap=None, zscale = None):
    """
    Create a surface plot from a 2D array using PyVista and PolyData.

    Parameters:
        data_2d (numpy.ndarray): A 2D numpy array to plot.
        plotter (pyvista.Plotter): Existing plotter instance (if any).
        show_edges (bool): Whether to show edges of the surface.
        cmap (str): Colormap to use for the surface.
    """
    import pyvista as pv
    
    if not isinstance(data_2d, np.ndarray) or len(data_2d.shape) != 2:
        raise ValueError("Input must be a 2D numpy array.")
    
    nx, ny = data_2d.shape
    x = np.arange(nx)
    y = np.arange(ny)
    yy, xx = np.meshgrid(y, x)
    z = data_2d
    points = np.c_[xx.ravel(), yy.ravel(), z.ravel()]
    faces = []

    for i in range(nx - 1):
        for j in range(ny - 1):
            idx = i * ny + j
            faces.append([4, idx, idx + 1, idx + ny + 1, idx + ny])

    faces = np.array(faces, dtype=np.int32).ravel()

    mesh = pv.PolyData(points, faces)
    mesh["scalars"] = z.ravel()
    if plotter is None:
        plotter = pv.Plotter()
    plotter.add_mesh(mesh, show_edges=show_edges, cmap=cmap, scalars="scalars",
                     scalar_bar_args={'title': "Intensity"})
    if zscale is None:
        zscale = (data_2d.shape[0] * data_2d.shape[1])**0.5 / (
            data_2d.max() - data_2d.min())
    plotter.set_scale(xscale = 1, yscale = 1, zscale = zscale)
    plotter.add_axes()
    
    return plotter